#!/usr/bin/python
import sklearn.metrics
import HTSeq
import numpy as np ; np.seterr(all = "ignore")



def mutual_information(VAR_LIST, var_type_1, var_type_2, SM,
    mi_testable_distance = 12000, mi_testable_common_reads = 10, 
    mi_testable_mono = 0, mi_testable_ratio = 0.9, intronic_alleles_dict = None):
    
    mi_info   = dict()

    var1_list = []
    
    for POS, ALLELE_dict in VAR_LIST.items():
        if var_type_1 in POS:
            for ALLELE, attr in ALLELE_dict.items():
                if SM in attr and attr[SM]["FEAT"]["FILTER"] == "PASS" :
                    var1_list.append((POS, ALLELE))

    var2_list = []
    
    for POS, ALLELE_dict in VAR_LIST.items():
        if var_type_2 in POS:
            for ALLELE, attr in ALLELE_dict.items():
                if SM in attr and attr[SM]["FEAT"]["FILTER"] == "PASS" :
                    var2_list.append((POS, ALLELE))

    var1_list.sort(key = lambda x : x[0][1])
    var2_list.sort(key = lambda x : x[0][1])


    for var1 in var1_list:
        var1_reads = VAR_LIST[var1[0]][var1[1]][SM]["READS"] 
        var1_id    = VAR_LIST[var1[0]][var1[1]]['variant_id']
        var1_pos   = var1[0][1]
        var1_end   = var1[1][2]
        
        MIs = []
        MI_n = []
        tag_vars = []

        for var2 in var2_list:
            var2_reads = VAR_LIST[var2[0]][var2[1]][SM]["READS"] 
            var2_id    = VAR_LIST[var2[0]][var2[1]]['variant_id']
            var2_pos   = var2[0][1]
            var2_end   = var2[1][2]
                   
            common_reads = set(var1_reads) & set(var2_reads)
            lc = len(common_reads)


            if lc < mi_testable_common_reads:
                continue
            elif var1_pos - var2_end > mi_testable_distance:
                continue
            elif var2_pos - var1_end > mi_testable_distance:
                break
            elif var1_pos == var2_pos:
                continue

            mat = np.zeros((lc, 2)).astype(int)
            for i, r in enumerate(common_reads):
                mat[i][0] = var1_reads[r]
                mat[i][1] = var2_reads[r]

            if intronic_alleles_dict is not None:
                umat, ucounts = np.unique(mat, axis = 0, return_counts = True)
                for u_i in range(umat.shape[0]):
                    (jxn_s, jxn_e) = intronic_alleles_dict[var2_id][umat[u_i][1]]   
                    jxn_mid = (jxn_e + jxn_s)//2
                    jxn_count = ucounts[u_i]
                    # outline = ["[MI Log sashimi]", var1_id, var2_id, umat[u_i][0], "{}-{}".format(jxn_s, jxn_e), jxn_count]
                    # print("\t".join(map(str, outline)))


            var1_alleles, var1_allele_counts = np.unique(mat[:, 0], return_counts = True)
            var2_alleles, var2_allele_counts = np.unique(mat[:, 1], return_counts = True)

            tag = "likely_heterozygous"

            for AC in var1_allele_counts:
                if AC > lc - mi_testable_mono or AC/lc > mi_testable_ratio:
                    tag = "likely_homozygous"

            for AC in var2_allele_counts:
                if AC > lc - mi_testable_mono or AC/lc > mi_testable_ratio:
                    tag = "likely_homozygous"

            if tag == 'likely_homozygous':
                continue 

            tag_vars.append((var2_id, var2_alleles))
            MI_score = sklearn.metrics.mutual_info_score(mat[:, 0], mat[:, 1])
            MIs.append(round(MI_score, 3))
            MI_n.append(lc)

        if MIs:
            n = len(MI_n)
            wAve_MI  = np.average(MIs, weights = MI_n)
            mi_info[var1] = [n, round(wAve_MI, 3), tag_vars, MIs, MI_n]
        else:
            mi_info[var1] = [0, np.nan, [], [], []]

    return mi_info




def mi_parse_variants(VAR_LIST, SM, options):

    printing_lines = []

    mi_info = mutual_information(VAR_LIST, "SNV", "SNV", SM, 
                mi_testable_common_reads = options.link_min_count,
                mi_testable_mono  = 2, 
                mi_testable_ratio = 0.9)

    for var, (n, mi, tag_vars, mi_vars, mi_n) in mi_info.items():
        (POS, ALLELE) = var

        VAR_LIST[POS][ALLELE][SM]["FEAT"]["MI"]   = mi
        VAR_LIST[POS][ALLELE][SM]["FEAT"]["MI_n"] = n

        if n:
            strand = VAR_LIST[POS][ALLELE]['STRAND']
            qual   = VAR_LIST[POS][ALLELE][SM]["FEAT"]['QUAL']
            var_id = VAR_LIST[POS][ALLELE]['variant_id']

            if mi < 0.3 and n >= 2:
                VAR_LIST[POS][ALLELE]['variant_type'] = "RNA_EDITING"
        
            mi_vars = ["{:.3f}".format(_mi) for _mi in mi_vars]
            
            line = [SM, "SNV", "SNV", var_id, qual, mi, n, ";".join([v[0] for v in tag_vars]), 
                        ";".join(mi_vars), ";".join(map(str, mi_n)), strand]
            
            printing_lines.append(line)


    
    mi_info = mutual_information(VAR_LIST, "INDEL", "SNV", SM,
                mi_testable_common_reads = options.link_min_count,
                mi_testable_mono  = 2, 
                mi_testable_ratio = 0.9)

    for var, (n, mi, tag_vars, mi_vars, mi_n) in mi_info.items():
        (POS, ALLELE) = var

        VAR_LIST[POS][ALLELE][SM]["FEAT"]["MI"]   = mi
        VAR_LIST[POS][ALLELE][SM]["FEAT"]["MI_n"] = n

        if n:
            strand = VAR_LIST[POS][ALLELE]['STRAND']
            qual   = VAR_LIST[POS][ALLELE][SM]["FEAT"]['QUAL']
            var_id = VAR_LIST[POS][ALLELE]['variant_id']

            mi_vars = ["{:.3f}".format(_mi) for _mi in mi_vars]
            
            line = [SM, "INDEL", "SNP", var_id, qual, mi, n, ";".join([v[0] for v in tag_vars]), 
                        ";".join(mi_vars), ";".join(map(str, mi_n)), strand]
            
            printing_lines.append(line)

    
    IP_alleles_dict = {}

    for POS, ALLELE_dict in VAR_LIST.items():
        if "INTRONIC_PART" in POS:
            for ALLELE, attr in ALLELE_dict.items():
                if SM in attr:
                    ip_introns = [(intron[0][1], intron[1][2]) for intron in attr[SM]["IP_ALLELES"]]
                    ip_var_id  = attr["variant_id"]
                    IP_alleles_dict[ip_var_id] = ip_introns


    for var1_type in ["INDEL", "SNV", "RNA_EDITING"]:
        
        mi_info = mutual_information(VAR_LIST, var1_type, "INTRONIC_PART", SM, 
                                        mi_testable_common_reads = options.link_min_count,
                                        mi_testable_mono  = 0, 
                                        mi_testable_ratio = 1.0,
                                        intronic_alleles_dict = IP_alleles_dict)


        for var, (n, mi, tag_vars, mi_vars, mi_n) in mi_info.items():
            (POS, ALLELE) = var

            VAR_LIST[POS][ALLELE][SM]["FEAT"]["MI"]   = mi
            VAR_LIST[POS][ALLELE][SM]["FEAT"]["MI_n"] = n

            strand = VAR_LIST[POS][ALLELE]['STRAND']
            qual   = VAR_LIST[POS][ALLELE][SM]["FEAT"]['QUAL']
            var_id = VAR_LIST[POS][ALLELE]['variant_id']
            
            for j, (ip, ip_alleles) in enumerate(tag_vars):
                mi, lc = mi_vars[j], mi_n[j]
                
                ip_alleles_string = ["SP{0}:{1[0]}-{1[1]}".format(i + 1, IP_alleles_dict[ip][i]) \
                                        for i in ip_alleles]

                ip_alleles_string = ip + ";".join(ip_alleles_string)
                
                line = [SM, var1_type, "INTRONIC_PART", var_id, qual, mi, 1, ip_alleles_string, 
                         mi, lc, strand]
                
                printing_lines.append(line)

    return printing_lines












#!/usr/bin/python

import numpy as np ; np.seterr(all = "ignore")
import pandas as pd
import matplotlib ; matplotlib.use('Agg')
import random
import _pickle as pickle
import pyfaidx
import matplotlib.pyplot as plt

from glob import glob
from optparse import OptionParser, OptionGroup
from collections import Counter, defaultdict

from sklearn.metrics         import roc_curve, auc, f1_score
from sklearn.model_selection import ShuffleSplit, GridSearchCV
from sklearn.linear_model    import LogisticRegression
from sklearn.utils           import resample

import sys
import os
import re
import vcf
from vcf.parser import _Info as VcfInfo





def update_vars(GLM_ins, GLM_del, GLM_snp, ALL_VARS_FEAT, SM, options):
	for POS, ALT_dict in ALL_VARS_FEAT.items():
		if "INTRON" in POS or "INTRONIC_PART" in POS:
			continue

		for ALLELE, var_dict in ALT_dict.items():
			if SM not in var_dict:
				continue
			else:
				qual = covert_to_Featurevector(var_dict, SM, GLM_ins, GLM_del, GLM_snp, options, )

			sm_feat = var_dict[SM]["FEAT"]
			
			ALL_VARS_FEAT[POS][ALLELE][SM]["FEAT"]["QUAL"] = qual

			### Filters
			if sm_feat['OVER_PLOIDY']:
				FILTER = 'Overploidy'
			elif sm_feat['AC'] < options.minCountVar:
				FILTER = 'LowAC'
			elif sm_feat['DP'] < options.minCoverage:
				FILTER = 'LowDP'
			elif sm_feat['AB'] < options.minRatioVar:
				FILTER = 'LowAB'
			elif sm_feat['BASEQUAL_ALT_mean'] < options.minBaseQual:
				FILTER = 'LowBaseQual'
			elif sm_feat['READPOS_ALT_mean']  < options.minReadPos:
				FILTER = 'LowReadPos'
			
			elif qual >= 3.01:
				FILTER = "PASS"
			else:
				FILTER = "LowQual"

	
			ALL_VARS_FEAT[POS][ALLELE][SM]["FEAT"]["FILTER"] = FILTER
		
			


def covert_to_Featurevector(var_dict, SM, model_INS, model_DEL, model_SNP, 
	options, pseudo_prob = 1e-10):

	sm_feat = var_dict[SM]["FEAT"]

	FV = [sm_feat["AB"], 
		  sm_feat['BASEQUAL_ALT_mean'], 
		  var_dict['REP_COUNT'], 
		  sm_feat["ndHAP"], 
		  sm_feat["NVARS"],
		  sm_feat["READPOS_ALT_mean"], 
		  np.log(sm_feat["pb"] + 1e-5 )]


	try:

		if var_dict["INDEL_LEN"] == 0:
			p_False, p_True = model_SNP.predict_proba([FV])[0]
		elif var_dict["INDEL_LEN"] > 0:
			p_False, p_True = model_INS.predict_proba([FV])[0]
		elif var_dict["INDEL_LEN"] < 0:
			p_False, p_True = model_DEL.predict_proba([FV])[0]

		QUAL = -10 * np.log10(p_False + pseudo_prob)

	except:
		QUAL = 0.0

	return round(QUAL, 2)



def variant_is_same(var_call, var_base, ref_genome, overhang = 25):
	base_chrom, base_pos, base_ref, base_alt = re.split('[:,>]', var_base)
	call_chrom, call_pos, call_ref, call_alt = re.split('[:,>]', var_call)
	
	base_pos, call_pos = int(base_pos), int(call_pos)
	
	if var_call == var_base:
		return True
	elif abs(base_pos - call_pos) > overhang:
		return False
	elif len(base_ref) * len(base_alt) * len(call_ref) * len(call_alt) == 1:
		return False
	elif (len(base_ref) - len(base_alt)) != (len(call_ref) - len(call_alt)):
		return False
	

	seq = ref_genome.fetch(call_chrom, base_pos - overhang, base_pos + overhang)
	seq = str(seq).upper()

	subseq = seq[overhang : overhang + len(base_ref)]

	assert subseq == base_ref.upper(), '{} != {}'.format(subseq, base_ref)
	
	ALT_BASE = seq[: overhang] + base_alt.upper() + seq[overhang + len(base_ref) : ]
	overhang = overhang + (call_pos - base_pos) 
	ALT_CALL = seq[: overhang] + call_alt.upper() + seq[overhang + len(call_ref) : ]

	if ALT_BASE == ALT_CALL:
		return True
	else:
		return False



def process_dataframe(gqv_training_file, baseline_vcf, ref_genome, overhang = 25):
	canonical_chroms = ["chr1","chr2", "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chr10",
						"chr11","chr12", "chr13","chr14","chr15","chr16","chr17","chr18","chr19","chr20",
						"chr21","chr22", "chrX"]

	df = pd.read_csv(gqv_training_file, index_col = None, sep = '\t')

	print("\tOriginal: rows=", len(df))

	df = df[(abs(df.INDEL_LEN) <= 20 )] 
	print("\tFiltering by length: rows=", len(df))

	df = df[(df.AC >= 2) & ( df.AB >= 0.05)] 
	print("\tFiltering by Cov: rows=", len(df))
	
	df = df[(df.BASEQUAL_ALT_mean >= 20)]
	print("\tFiltering by BASEQUAL: rows=", len(df))

	df = df[(df.READPOS_ALT_mean >= 7)]
	print("\tFiltering by READPOS: rows=", len(df))

	df = df[(df.variant_type != "RNA_EDITING?")]
	print("\tFiltering by editing type rows=", len(df))

	df = df[(df.DP <= 10)]
	print("\tFiltering by coverage =", len(df))


	df = df.fillna(0)
	
	df['pb'] = np.log(df['pb'] + 1e-5)

	vcf_file_handle = vcf.Reader(filename = baseline_vcf)

	vcf_file_handle.infos['datasetsmissingcall'] = VcfInfo('datasetsmissingcall', '.', 'String', "bar", 'bar', 'bar')
	vcf_file_handle.infos['testable']            = VcfInfo('testable', '.', 'String', "bar", 'bar', 'foo')
	
	Y_alt = []
	var_ids = []

	for i, row in df.iterrows():
		CHROM, start, end, REF, ALT = re.split('[-,:,>]', row['variant_id'])
		start, end = int(start) + 1, int(end)

		var_id = f'{CHROM}:{start}:{REF}>{ALT}'

		found = False 
		label = False

		if CHROM in canonical_chroms:

			for record in vcf_file_handle.fetch(CHROM, start - overhang, end + overhang):
				for record_ALT in record.ALT:
					var_id_base = f"{CHROM}:{record.POS}:{record.REF}>{record_ALT}"
					label = variant_is_same(var_id, var_id_base, ref_genome)
					# print("checking", row["REP_COUNT"], row["variant_type"], var_id, var_id_base, label )
					if label: 
						found = True

				if found:
					break

		if i % 50000 == 0:
			sys.stderr.write("\tLine counter {}...\n".format(i))
			sys.stderr.flush()

		var_ids.append(var_id)
		Y_alt.append(int(label))


	df["LABEL"]  = Y_alt

	return df 




def glm_modelling(gqv_training_file, df):
	#

	Feats = ['AB', 'BASEQUAL_ALT_mean', "REP_COUNT", 'ndHAP', "NVARS", "READPOS_ALT_mean", "pb"]

	for VAR_TYPE, (l1, l2) in [("INSERTION", (1, 20)), ("DELETION", (-21, -1)), ("SNP", (0, 0))]:

		X = df[( df.INDEL_LEN >= l1 ) & ( df.INDEL_LEN <= l2) ] 
		X = X.reset_index(drop = True)
		Y = X.LABEL 
		X = X[Feats] 

		sys.stderr.write('\n\nVARIANT = {}\tLabel count: TP={} FP={}\n'.format(VAR_TYPE, sum(Y), len(Y) - sum(Y)))

		scores = list()
		fig = plt.figure()
		ax = fig.add_subplot(111)

		splits = ShuffleSplit(n_splits = 20, test_size = 0.33, random_state = 0)
		splits.get_n_splits(X, Y)

		LogRegClassifier = LogisticRegression(solver = 'liblinear', class_weight = 'balanced')

		for (train, test) in splits.split(X):
			X_train, Y_train = X.loc[train], Y[train]
			X_test,  Y_test  = X.loc[test],  Y[test]

			probas = LogRegClassifier.fit(X_train, Y_train).predict_proba(X_test)
			fpr, tpr, thresholds = roc_curve(Y_test, probas[:, 1])
			roc_auc = auc(fpr, tpr)

			Y_pred = LogRegClassifier.fit(X_train, Y_train).predict(X_test)
			f1_s = f1_score(Y_test, Y_pred)

			scores.append((f1_s, roc_auc))

			ax.plot(fpr, tpr, lw = 0.6)

		ax.plot([0, 1], [0, 1], transform = ax.transAxes, color='r', ls='--')
		ax.set_xlim([-0.05, 1.05])
		ax.set_ylim([-0.05, 1.05])
		ax.set_xlabel('False Positive Rate')
		ax.set_ylabel('True Positive Rate')

		mean_f1, mean_auroc = np.mean(scores, axis = 0)
		std_f1, std_auroc   = np.std(scores, axis = 0)

		Title1 = 'ROC GLM classifier for micro-indels (training set = {})\n'.format(gqv_training_file)
		Title2 = 'Mean F1 = {:.2f}, Mean AUROC = {:.2f}\n'.format(mean_f1, mean_auroc)
		Title3 = 'Std. Dev. F1 = {:.3f}, Std. Dev. AUROC = {:.3f}'.format(std_f1, std_auroc)

		sys.stderr.write('\n\t' + Title2 + '\t\t' + Title3 + '\n')

		ax.set_title(Title1 + Title2 + Title3)

		fig.savefig("{}.ROC_for_training_set.Var={}.pdf".format(gqv_training_file, VAR_TYPE))

		pickle_out_name = '{}.{}.glm.pickle'.format(gqv_training_file, VAR_TYPE)

		tmp_feat_vals = dict((k, round(v, 2)) for k, v  in zip(Feats, LogRegClassifier.coef_[0]))

		sys.stderr.write("\tcoefficients: {} intercept: {:.2f}\n".format(tmp_feat_vals, LogRegClassifier.intercept_[0]))
		sys.stderr.write("\n\tWriting output to {}\n".format(pickle_out_name))

		with open(pickle_out_name, 'wb') as fid:
			pickle.dump(LogRegClassifier, fid)

		ps = LogRegClassifier.predict_proba(df[Feats])
		p_False, p_True = ps[:,0], ps[:,1]
		df['QUAL_{}'.format(VAR_TYPE)] = -10 * np.log10(p_False + 1e-10)	
	
	df.to_csv(gqv_training_file + ".proc.csv", index = False, sep = "\t")



def main():
	args = sys.argv

	description  = """scAllele's training module:\n"""
	description += """With a feature matrix file (run scAllele in "Training" mode)\n"""
	description += """and a groud truth set of variants, train a new variant classifier.\n"""

	usage = """\n\tscAllele_train -i <Feat_Matrix.tab> -v <truth.vcf> -g <genome.fa> """

	parser = OptionParser(usage = usage, description = description)
	
	parser.add_option("-i", 
		dest = "featmat", 
		help = "[Required] scAllele's Feature Matrix file")
	parser.add_option("-v",     
		dest = "truth_vcf", 
		help = "[Required] Groundtruth set of variants (vcf format)")
	parser.add_option("-g",  
		dest = "Genome", 
		help = "[Required] Reference Genome (fasta format)")

	(options, Args) = parser.parse_args()

	if len(sys.argv) <= 4:
		parser.print_help()
		sys.exit(1)

	ref_file = pyfaidx.Faidx(options.Genome)

	df = process_dataframe(options.featmat, options.truth_vcf, ref_file)

	glm_modelling(options.featmat,  df)



if __name__ == '__main__':
	main()




#!/usr/bin/python
import pysam
from collections import defaultdict, Counter
from HTSeq import GenomicInterval
from time import time
import itertools
import random
import numpy as np ; np.seterr(all = "ignore")
import os 


def read_bam_input(option_bam):
	if "," in option_bam:
		bam_list = option_bam.split(',')
	elif option_bam.endswith('.bam'):
		bam_list = [option_bam]
	else:
		bam_list = []
		with open(option_bam) as f:
			for line in f:
				bam_file = line.strip() 
				bam_list.append(bam_file)

	for bam_file in bam_list:
		assert os.path.exists(bam_file) & os.path.exists(bam_file + '.bai'), f"invalid bam {bam_file}"

	return bam_list



def filter_reads(pysam_read, P_minMapQual):
	tags = dict(pysam_read.tags)

	if pysam_read.mapping_quality < P_minMapQual:
		return True, 'mapq'
	elif len(pysam_read.query_alignment_sequence) < 15:
		return True, 'short'
	elif "N" in pysam_read.query_alignment_sequence:
		return True, 'N'
	elif pysam_read.is_unmapped:
		return True, 'Unmap'
	elif pysam_read.is_secondary:
		return True, 'Second'
	elif pysam_read.is_duplicate:
		return True, 'Dup'
	elif pysam_read.is_qcfail:
		return True, 'qcfail'
	elif "NH" in tags and tags["NH"] > 1:
		return True, 'nh>1'
	elif pysam_read.is_paired and not pysam_read.is_proper_pair:
		return True, 'unpair'
	elif pysam_read.is_supplementary:
		return True, 'supp'
	else:
		return False, 'good'



def get_read_strand(pysam_read, strandedness):
	
	if strandedness == 'fr-firststrand':
		strand_1, strand_2 = '+', '-'
	elif strandedness == 'fr-secondstrand':
		strand_1, strand_2 = '-', '+'
	elif strandedness == 'fr-unstrand':
		strand_1, strand_2 = '.', '.'

	if pysam_read.is_read2 and pysam_read.is_reverse:      
		r_strand = strand_2			
	elif pysam_read.is_read1 and pysam_read.is_reverse:	     
		r_strand = strand_1
	elif pysam_read.is_read1 and not pysam_read.is_reverse:  
		r_strand = strand_2
	elif pysam_read.is_read2 and not pysam_read.is_reverse:  
		r_strand = strand_1
	else:
		if pysam_read.is_reverse:
			r_strand = strand_1
		else:
			r_strand = strand_2
	return r_strand


def process_CIGAR(pos, cigar):
	block_lengths  = [] 
	soft_clippings = []
	blocks = 0
	for (ctype, length) in cigar:
		if ctype == 0: #M
			blocks += length
		elif ctype == 2: #D
			blocks += length
		elif ctype == 3: #N
			block_lengths.append((pos, pos + blocks))
			pos += (blocks + length)
			blocks = 0
		elif ctype == 4: #S
			soft_clippings.append(length)
	block_lengths.append((pos, pos + blocks))
	return block_lengths, soft_clippings


def get_chroms_from_bam(bam_file):
	bam_handle = pysam.Samfile(bam_file, 'rb')
	bam_chroms = bam_handle.references
	bam_handle.close()
	return bam_chroms



def get_sample_name(bam_file):
	bam_handle = pysam.Samfile(bam_file, 'rb')	
	try:
		sample_name = bam_handle.header['RG'][0]['SM']
	except:
		sample_name = bam_file
	
	return sample_name



def find_super_read_clusters(bam_file_list, options, chrom = None, f_start = None, f_end = None):
	read_block_coords = defaultdict(dict)

	for bam_file in bam_file_list:

		bam_handle = pysam.Samfile(bam_file, 'rb')

		for read in bam_handle.fetch(chrom, start = f_start, end = f_end):
			chrom = read.reference_name

			if filter_reads(read, options.minMapQ)[0]:
				continue

			sB, eB = read.positions[0], read.positions[-1] 

			try:
				read_block_coords[chrom][sB] += 1
			except:
				read_block_coords[chrom][sB] = 1
			try:
				read_block_coords[chrom][eB] -= 1
			except:
				read_block_coords[chrom][eB] = -1

		bam_handle.close()


	Read_Clusters = []
	RC_cov_cutoff = options.minCoverage 

	for chrom, coords in read_block_coords.items():
		Cum_Cov_block = 0
		cluster_count = 0 

		RC_started = False

		coords = sorted(coords)

		for coord in coords: 
			Cum_Cov_block += read_block_coords[chrom][coord]
		
			if Cum_Cov_block >= RC_cov_cutoff and not RC_started:
				start_coord = coord
				RC_started  = True
			elif Cum_Cov_block < RC_cov_cutoff and RC_started:
				cluster_count += 1
				if cluster_count >= 1000 or coord == coords[-1]:
					Read_Clusters.append([chrom, start_coord, coord])
					RC_started = False
					cluster_count = 0 

	return Read_Clusters



def find_read_clusters(bam_file, options, chrom, start, end):
	
	read_block_coords = defaultdict(dict)

	bam_handle = pysam.Samfile(bam_file, 'rb')

	for read in bam_handle.fetch(chrom, start = start, end = end, until_eof = True):
		chrom = read.reference_name

		if filter_reads(read, options.minMapQ)[0]:
			continue

		r_blocks, soft_clippings = process_CIGAR(read.pos, read.cigar)

		if any(SC > options.maxSoftClip for SC in soft_clippings):
			continue

		r_strand = get_read_strand(read, options.strandedness)

		new_read = remove_homopolymer_ends(read, r_blocks)
		
		if new_read is None:
			continue

		aligned_seq, aligned_qual, r_blocks = new_read


		for i, (sB, eB) in enumerate(r_blocks):
			try:
				read_block_coords[(chrom, r_strand)][sB] += 1
			except:
				read_block_coords[(chrom, r_strand)][sB] = 1
			try:
				read_block_coords[(chrom, r_strand)][eB] -= 1
			except:
				read_block_coords[(chrom, r_strand)][eB] = -1

	bam_handle.close()


	Read_Clusters = list()
	RC_cov_cutoff = options.minCoverage 
	i = 0

	for (chrom, strand), coords in read_block_coords.items():
		Cum_Cov_block = 0
		Max_Cov_block = 0

		RC_started = False
		prev_coord = 0
		Area = 0.0

		_read_clusters = list()

		for coord in sorted(coords): 
			Cum_Cov_block += read_block_coords[(chrom, strand)][coord]
		
			if Cum_Cov_block > Max_Cov_block:
				Max_Cov_block = Cum_Cov_block

			if Cum_Cov_block >= RC_cov_cutoff and not RC_started:
				start_coord = coord
				Area = 0.0
				RC_started = True
				

			elif Cum_Cov_block < RC_cov_cutoff and RC_started:
				dx = coord - prev_coord
				dy = Cum_Cov_block  
				Area += dx*dy 

				_read_clusters.append((start_coord, coord, Max_Cov_block, Area))
				RC_started = False
				Max_Cov_block = 0
				Area = 0

			else:
				dx = coord - prev_coord
				dy = Cum_Cov_block  
				Area += dx*dy 

			prev_coord = coord 

		for (start, end, Max_Cov, Area) in _read_clusters:
			RC_length = end - start		

			v = (end - start)*Max_Cov
			
			if Max_Cov < options.minCoverage:
				continue

			if RC_length > 5000:
				sub_RC_length = 3000
				remainder = RC_length % sub_RC_length
				while remainder < 350 and remainder > 0:
					sub_RC_length += 5
					remainder = RC_length % sub_RC_length

				for s_start in range(start, end, sub_RC_length):
					s_end = min(s_start + sub_RC_length, end)
					RC = [i, chrom, strand, s_start, s_end, Max_Cov]
					Read_Clusters.append(RC)
					i += 1
			else:
				RC = [i, chrom, strand, start, end, Max_Cov]
				Read_Clusters.append(RC)
				i += 1

		del _read_clusters
		
	del read_block_coords

	return Read_Clusters


	
adapter1, adapter2 = "XX", "YY"


class genome_ref_pos():
	
	def __init__(self, CHROM, Covered_Ranges, RC_Range, RC_index, strand, genome):

		RC_ext_Start = Covered_Ranges[0][0] 
		RC_ext_End   = Covered_Ranges[-1][1]

		RC_ext_seq_compressed = adapter1
		for (s, e) in Covered_Ranges:
			block_seq = str(genome.fetch(CHROM, s + 1, e)).upper()
			RC_ext_seq_compressed += block_seq  + "I"
		RC_ext_seq_compressed = RC_ext_seq_compressed.rstrip("I")	
		RC_ext_seq_compressed += adapter2

		RC_ext_seq_full  = adapter1 
		RC_ext_seq_full += str(genome.fetch(CHROM, RC_ext_Start + 1, RC_ext_End)).upper()
		RC_ext_seq_full += adapter2

		self.Seq_ext      = RC_ext_seq_compressed
		self.Seq_ext_full = RC_ext_seq_full 

		RC_Start, RC_End  = RC_Range
 
		self.gi     = GenomicInterval(CHROM, RC_Start, RC_End)
		self.ext_gi = GenomicInterval(CHROM, RC_ext_Start, RC_ext_End)

		self.ranges = Covered_Ranges
		self.strand = strand 
		self.RC_i   = RC_index

	def rc_pos(self, genome_pos):
		RC_Pos = None
		CumLen = 2
		for (s, e) in self.ranges:
			if s <= genome_pos and genome_pos < e:
				RC_Pos = CumLen + genome_pos - s
			CumLen += (e - s)  + 1
		return RC_Pos

	def get_sequence(self, g_start, g_end):
		off = g_start - self.ext_gi.start + 2
		assert off >= 0 , f"VAR start {g_start} off {off} GI {self.gi} GI_ext {self.ext_gi}"
		return self.Seq_ext_full[off : off + (g_end - g_start)]

	def genome_pos(self, relative_pos):
		CumLen = len(adapter1)
		RealPos = None
		for (s, e) in self.ranges:
			ExonLen = e - s 
			if relative_pos > CumLen + ExonLen:
				CumLen += ExonLen + 1
			else:
				RealPos = s + relative_pos - CumLen
				break
		if RealPos is None:
			RealPos = self.ext_gi.end
		return RealPos


def remove_homopolymer_ends(read, blocks):
	READSEQ = read.query_alignment_sequence
	if "A"*50 in READSEQ or "C"*50 in READSEQ or "G"*50 in READSEQ or "T"*50 in READSEQ:
		return None
	i = 0
	lim = len(READSEQ) - 12
	while i < lim and READSEQ[i].upper() == READSEQ[i + 1].upper():
		i += 1

	j = len(READSEQ)
	lim = 12
	while j > lim and READSEQ[j - 1].upper() == READSEQ[j - 2].upper():
		j -= 1

	prefix = i if i > 10 else 0
	sufix  = j if j < (len(READSEQ) - 10) else len(READSEQ)

	if sufix - prefix <= 15:
		prefix, sufix = 0, len(READSEQ)

	new_blocks = []
	rlen = 0
	for (b1, b2) in blocks:
		blen = (b2 - b1)
		b1_new, b2_new = b1, b2

		if rlen + blen <= prefix:
			rlen += blen
			continue
		else:
			b1_new = b1 + max(prefix - rlen, 0)

		if rlen + blen <= sufix:
			pass
		elif rlen > sufix:
			continue
		else:
			b2_new = b1 + (sufix - rlen)

		rlen += blen
		new_blocks.append((b1_new, b2_new))

	aligned_qual = read.query_alignment_qualities[prefix : sufix]
	aligned_seq  = read.query_alignment_sequence[prefix : sufix]	

	return aligned_seq, aligned_qual, new_blocks



def check_read_pileup(SetOfReadNames):

	## read mono-clusters

	if any(len(v['blocks']) != 1 for v in SetOfReadNames.values()):
		return 0.0

	if len(SetOfReadNames) > 300:
		r_names = random.sample(list(SetOfReadNames), 300)
	else:
		r_names = list(SetOfReadNames)

	read_pairwise_overlap = []
	for (r1, r2) in itertools.combinations(r_names, 2):
		(r1_bs, r1_be) = SetOfReadNames[r1]['blocks'][0]
		(r2_bs, r2_be) = SetOfReadNames[r2]['blocks'][0]
		r1_len = r1_be - r1_bs
		r2_len = r2_be - r2_bs
		
		if r1_bs <= r2_be and r2_bs <= r1_be:
			ov = min([r2_len, r1_len, r1_be - r2_bs, r2_be - r1_bs])
		else:
			ov = 0.0
		ov = ov/min(r1_len, r2_len)
		read_pairwise_overlap.append(ov)

	RC_read_overlap = np.median(read_pairwise_overlap) 
	return RC_read_overlap 



def get_RC_reads(bam_file, RC_info, genome, SM, options):

	RC_index, chrom, strand, RC_Start, RC_End, MaxCov = RC_info

	SetOfReadNames = defaultdict(dict)	

	_read_blocks = defaultdict(int)
	_unique_seqs = defaultdict(dict)
	_unique_pos  = defaultdict(dict)


	subsample_rate = 2000/float(MaxCov) 

	bam_handle = pysam.Samfile(bam_file, 'rb')

	random.seed(0)
	
	for ri, read in enumerate(bam_handle.fetch(chrom, RC_Start, RC_End)):	

		if random.random() > subsample_rate:
			continue

		read_name = "{}:{}:{}".format(SM, read.query_name, read.is_read1)

		if get_read_strand(read, options.strandedness) != strand:
			continue

		if filter_reads(read, options.minMapQ)[0]:
			continue
		
		blocks, soft_clippings = process_CIGAR(read.pos, read.cigar)

		if not any(b1 <= RC_End and RC_Start <= b2 for (b1, b2) in blocks):
			continue

		if any(SC > options.maxSoftClip for SC in soft_clippings):
			continue

		new_read = remove_homopolymer_ends(read, blocks)
		if new_read is None:
			continue
		aligned_seq, aligned_qual, new_blocks = new_read
		for (b1, b2) in new_blocks:
			_read_blocks[b1] += 1
			_read_blocks[b2] -= 1

		SetOfReadNames[read_name]['Dir']    = read.is_reverse
		SetOfReadNames[read_name]['Quals']  = aligned_qual
		SetOfReadNames[read_name]['seq']    = aligned_seq
		SetOfReadNames[read_name]['blocks'] = new_blocks
		SetOfReadNames[read_name]['PcrDup'] = read.is_duplicate
		
		_unique_pos[(tuple(new_blocks), aligned_seq)][read_name] = sum(aligned_qual)


	bam_handle.close()

	RC_overlap_median = check_read_pileup(SetOfReadNames)

	if RC_overlap_median >= 0.95:
		return None


	SetOfReadGroups = defaultdict(dict)
	list_index = 1

	for (unique_block, unique_seq), read_dict in _unique_pos.items():
		SetOfReadGroups[list_index]['seq']    = unique_seq
		SetOfReadGroups[list_index]['blocks'] = unique_block
		SetOfReadGroups[list_index]['reads']  = list(read_dict.keys())
		list_index += 1
		


	RC_cov_cutoff = 1
	Cum_Cov    = 0
	RC_started = False
	Covered_Ranges = []
	for coord in sorted(_read_blocks.keys()):
		Cum_Cov += _read_blocks[coord]

		if Cum_Cov >= RC_cov_cutoff and not RC_started:
			start_coord = coord
			RC_started = True
		elif Cum_Cov < RC_cov_cutoff and RC_started:
			if Covered_Ranges:
				if start_coord - Covered_Ranges[-1][1] > 20:
					Covered_Ranges.append([start_coord, coord])
				else:
					Covered_Ranges[-1][1] = coord
			else:
				Covered_Ranges.append([start_coord, coord])
			RC_started = False

	del _read_blocks, _unique_seqs
	
	if not Covered_Ranges or sum((b - a) for a, b in Covered_Ranges) < options.kmer*2:
		return None
	else:
		REFPOS = genome_ref_pos(chrom, Covered_Ranges, (RC_Start, RC_End), RC_index, strand, genome)
		Ref_Blocks = [(REFPOS.ext_gi.start - len(adapter1), REFPOS.ext_gi.end + len(adapter2))]
		
		SetOfReadNames[0]['Blocks'] = Ref_Blocks		
		SetOfReadGroups[0] = {'seq': REFPOS.Seq_ext, 'reads': [0], 'blocks': Ref_Blocks}

		return SetOfReadNames, REFPOS, SetOfReadGroups






#!/usr/bin/python
import re
import sys
import numpy as np ; np.seterr(all = "ignore")
import itertools 
import HTSeq
from time import time
from collections import defaultdict, Hashable
from Bio import pairwise2
import functools

sys.setrecursionlimit(100000)

## General scoring

_Intron_Gap = 500

_Gap = 5

_min_complex_length = 4

## Short sequence scoring

_gap_open, _gap_ext = -0.6*_Gap, -0.4*_Gap  

_base_pair_score = {}

bases = ["A", "C", "G", "T", "I", "N"]

for b1, b2 in itertools.product(bases, bases):
	if b1 == b2:
		_base_pair_score[(b1, b2)] = 0.0
	elif b1 == "I" or b2 == "I":
		_base_pair_score[(b1, b2)] = -1*_Intron_Gap
	elif b1 == "N" or b2 == "N":
		_base_pair_score[(b1, b2)] = -1
	else:
		_base_pair_score[(b1, b2)] = -1




class memoized(object):
	def __init__(self, function_name):
		self.function_name = function_name
		self.cache = {}

	def __call__(self, *args):
		if not isinstance(args, Hashable):
			return self.function_name(*args)
		if args in self.cache:
			return self.cache[args]
		else:
			value = self.function_name(*args)
			self.cache[args] = value
			return value

	def __get__(self, obj, objtype):
		return functools.partial(self.__call__, obj)


class variant:
	def __init__(self, g_start, g_end, REF, ALT, r_pos):
		self.g_start = g_start
		self.g_end   = g_end
		self.REF = REF.upper()
		self.ALT = ALT.upper()
		self.RIPos = r_pos
		self.RGPos = {}
		self.match_intron = False
		self.rep_count  = 0
		self.rep_region = None
    
	def __hash__(self):
		return hash(str(self))

	def __eq__(self, other):
		return str(self) == str(other)

	def L_ref(self): 
		return len(self.REF)

	def L_alt(self): 
		return len(self.ALT)

	def __str__(self):
		if self.L_ref() > 20:
			REFSEQ = "{}__({})__{}".format(self.REF[:10], self.L_ref(), self.REF[-10:])
		else:
			REFSEQ = self.REF
		return "{}-{}:{}>{}".format(self.g_start, self.g_end, REFSEQ, self.ALT)

	def length(self):
		return self.L_alt() - self.L_ref()

	def shift_Rpos(self, offset):
		for ri, (ri_s, ri_e) in self.RIPos.items():
			self.RIPos[ri] = (ri_s + offset, ri_e + offset)

	def get_variant_type(self):
		if   self.L_ref() == 0 and self.L_alt() == 0:
			variant_type = "NO_VARIANT"
		elif self.L_ref() == 1 and self.L_alt() == 1:
			variant_type = "SNV"
		elif self.L_ref() == 0 and self.L_alt() > 0:
			variant_type = "INSERTION"
		elif self.L_ref() > 0 and self.L_alt() == 0:
			if self.match_intron:
				variant_type = "INTRON"
			else:
				variant_type = "DELETION"
		elif self.L_ref() * self.L_alt() > 1:
			if self.REF[0] == self.ALT[0]:
				if self.L_ref() == 1 and self.L_alt() > 1:
					variant_type = "INSERTION"
				elif self.L_ref() > 1 and self.L_alt() == 1:
					variant_type = "DELETION"
				else:
					variant_type = "COMPLEX"
			else:
				variant_type = "COMPLEX"
		return variant_type


	def trim(self, first = "sufix", min_match = 3):
		Pfx, Sfx, ALT, REF = trim_sequences(self.ALT, self.REF, first, min_match)
		self.REF = REF 
		self.ALT = ALT
		self.g_start = self.g_start + Pfx
		self.g_end   = self.g_end - Sfx
		for read_name, (r_start, r_end) in self.RIPos.items():
			self.RIPos[read_name] = (r_start + Pfx, r_end - Sfx)		


	def same_variant(self, var2, REFPOS):
		min_s = min(self.g_start, var2.g_start)
		max_e = max(self.g_end, var2.g_end)
		hap1 = REFPOS.get_sequence(min_s, self.g_start) + self.ALT + REFPOS.get_sequence(self.g_end, max_e)
		hap2 = REFPOS.get_sequence(min_s, var2.g_start) + var2.ALT + REFPOS.get_sequence(var2.g_end, max_e)
		return hap1 == hap2


	def leftalign(self, REFPOS, max_shift = 100):
		max_shift = max(REFPOS.ext_gi.start + 1, self.g_start - max_shift)
		seq = REFPOS.get_sequence(max_shift, self.g_start + len(self.REF))
		assert seq.endswith(self.REF)

		if self.L_ref():
			seq = seq[:-self.L_ref()]
		
		ref, alt = self.REF, self.ALT 
		offset = 0
		quit = False
		while offset < len(seq) and not quit:
			quit = True
			if (ref and alt) and ref[-1] == alt[-1]:
				ref, alt = ref[:-1], alt[:-1]
				quit = False
			if len(ref) == 0 or len(alt) == 0:
				offset += 1
				ref = seq[-offset] + ref
				alt = seq[-offset] + alt
				quit = False

		self.REF = ref.upper()
		self.ALT = alt.upper() 
		self.g_start = self.g_start - offset
		self.g_end   = self.g_end - offset + 1
		for RI, (r_start, r_end) in self.RIPos.items():
			self.RIPos[RI] = (r_start - offset, r_end - offset + 1)


@memoized
def trim_sequences(rid_seq, ref_seq, first = 'prefix', min_match = 3):
	'''	
	Trim common sequences and label the variant.
	'''
	def _find_prefix(rid_seq, ref_seq):
		minLength = min(len(rid_seq), len(ref_seq))
		prefix = 0
		while prefix + 1 <= minLength and rid_seq[:prefix + 1] == ref_seq[:prefix + 1]:
			prefix += 1
		return prefix
	
	def _find_sufix(rid_seq, ref_seq):
		minLength = min(len(rid_seq), len(ref_seq))
		sufix = 0
		while sufix + 1 <= minLength and rid_seq[-(sufix + 1):] == ref_seq[-(sufix + 1):]:
			sufix += 1   
		return sufix

	sufix_final  = 0
	prefix_final = 0

	if first == 'prefix':

		prefix = _find_prefix(rid_seq, ref_seq)

		if prefix >= min_match or prefix == min(len(rid_seq), len(ref_seq)):
			rid_seq = rid_seq[prefix:]
			ref_seq = ref_seq[prefix:]
			prefix_final = prefix
		
		sufix = _find_sufix(rid_seq, ref_seq)		

		if sufix >= min_match or sufix == min(len(rid_seq), len(ref_seq)):
			rid_seq = rid_seq[:(-sufix or None)]
			ref_seq = ref_seq[:(-sufix or None)]
			sufix_final = sufix
	else:

		sufix = _find_sufix(rid_seq, ref_seq)

		if sufix >= min_match or sufix == min(len(rid_seq), len(ref_seq)):
			rid_seq = rid_seq[:(-sufix or None)]
			ref_seq = ref_seq[:(-sufix or None)]
			sufix_final = sufix

		prefix = _find_prefix(rid_seq, ref_seq)
		
		if prefix >= min_match or prefix == min(len(rid_seq), len(ref_seq)):
			rid_seq = rid_seq[prefix:]
			ref_seq = ref_seq[prefix:]
			prefix_final = prefix

	return prefix_final, sufix_final, rid_seq, ref_seq



def convert_to_var(REF, ALT, Ref_S, Ref_E, Rid_S, Rid_E, RI, REFPOS):

	Ref_S = REFPOS.genome_pos(Ref_S)
	Ref_E = REFPOS.genome_pos(Ref_E)
	if "I" in REF:
		REF = REFPOS.get_sequence(Ref_S, Ref_E)
	var = variant(Ref_S, Ref_E, REF, ALT, {RI : (Rid_S, Rid_E)})
	return var



def calculate_edit_distance(var, Junctions, REFPOS):

	is_intron = False 
	for (jxn_s, jxn_e) in sorted(Junctions):
		if jxn_s <= var.g_end and var.g_start <= jxn_e:
			jxn_var = variant(jxn_s, jxn_e, '', '', {})
			is_intron = var.same_variant(jxn_var, REFPOS)

	var.match_intron = is_intron

	if var.get_variant_type() == "INTRON":
		editd = _Intron_Gap
		var_set = [var]

	elif var.get_variant_type() == "NO_VARIANT":
		editd = 0.0 
		var_set = []

	elif var.get_variant_type() == "DELETION":
		editd = var.L_ref()*_Gap
		var_set = [var]

	elif var.get_variant_type() == "INSERTION":
		editd = var.L_alt()*_Gap
		var_set = [var]

	elif var.get_variant_type() == "SNV":
		editd = abs(_base_pair_score[(var.ALT, var.REF)])
		var_set = [var]

	elif var.get_variant_type() == "COMPLEX":
		editd, var_set = split_complex_var(var, Junctions, REFPOS)

	editd_wo_intron = editd
	n_vars = 0

	for v in var_set:
		if v.match_intron:
			editd_wo_intron -= _Intron_Gap
		if v.get_variant_type() in ('INSERTION', 'DELETION'):
			n_vars += 1
		elif v.get_variant_type() == 'COMPLEX':
			n_vars += 2

	return editd, editd_wo_intron, var_set, n_vars




def split_complex_var(var, Junctions, REFPOS):

	Matrix_Product_Cutoff_1 = 2000
	LenSim = np.log2(var.L_ref()/var.L_alt())

	if var.L_alt() <= _min_complex_length and var.L_ref() <= _min_complex_length:
		complex_var_type = "BOTH_SHORT"
		if var.L_ref() == var.L_alt():
			var_set, editd = edit_distance_function_msnp(var)
		else:
			editd = edit_distance_function_WA(var)
			var_set  = [var]

	elif var.L_alt() * var.L_ref() < Matrix_Product_Cutoff_1 and abs(LenSim) < 0.6:
		complex_var_type = "BOTH_MEDIUM"
		var_set, editd = edit_distance_function_biopython(var, Junctions, REFPOS)

	else:
		complex_var_type = "LONG"
		var_set, editd = edit_distance_function_internal_aln(var, Junctions, REFPOS)

	return editd, var_set



def edit_distance_lazy_calc(var):
	delta  = abs(var.L_ref() - var.L_alt())*_Gap
	common = min(var.L_ref(), var.L_alt())
	return common + delta



def edit_distance_function_msnp(var):
	editd = 0.0
	var_set = []
	for i in range(var.L_ref()):
		snp_ReadPos = {}
		for r_name, (r_start, r_end) in var.RIPos.items():
			snp_ReadPos[r_name] = (r_start + i, r_start + i + 1)
		snp_var = variant(var.g_start + i, var.g_start + i + 1, var.REF[i], var.ALT[i], snp_ReadPos)
		editd  += abs(_base_pair_score[(var.REF[i], var.ALT[i])])
		var_set.append(snp_var)
	return var_set, editd



def edit_distance_function_biopython(var, Junctions, REFPOS):
	aligner = pairwise2.align.globalds

	alignments = aligner(var.ALT, var.REF, _base_pair_score, _gap_open, _gap_ext)

	ALT_aln, REF_aln, Edit_Distance, Start, End = alignments[0]

	total_editd = 0
	coord_list = []

	ref_indexes, rid_indexes = [], []
	ref_i, rid_i = var.g_start, 0

	for (b_f, b_r) in zip(REF_aln + 'X', ALT_aln + 'X'):
		ref_indexes.append(ref_i)
		rid_indexes.append(rid_i)
		if b_f != "-" and b_r != "-":
			ref_i += 1
			rid_i += 1
		elif b_f == "-":
			rid_i += 1
		elif b_r == "-":
			ref_i += 1

	IN_started, DL_started = False, False

	for i, (b_f, b_r) in enumerate(zip(REF_aln, ALT_aln)):
		if b_f == "-" and b_r == "-":
			raise ValueError("Weird alignment {} {}\n".format(REF_aln, ALT_aln))
		if b_f == "-" and not IN_started:
			var_start = i
			IN_started = True
		elif b_f != "-" and IN_started:
			coord_list.append((var_start, i))
			IN_started = False
		if b_r == "-" and not DL_started:
			var_start = i
			DL_started = True
		elif b_r != "-" and DL_started:
			coord_list.append((var_start, i))
			DL_started = False
		if b_r != b_f and b_f != "-" and b_r != "-":
			coord_list.append((i, i + 1))
			 
	if IN_started or DL_started:
		coord_list.append((var_start, i + 1))

	final_var_set = []
	total_n_vars = 0
	for (s, e) in coord_list:
		tmp_Rid_S, tmp_Rid_E = rid_indexes[s], rid_indexes[e]
		tmp_Ref_S, tmp_Ref_E = ref_indexes[s], ref_indexes[e]
		ALT = ALT_aln[s : e].strip("-")
		REF = REF_aln[s : e].strip("-")
		_ReadPos = {}
		for r_name, (r_start, r_end) in var.RIPos.items():
			_ReadPos[r_name] = (r_start + tmp_Rid_S, r_start + tmp_Rid_E)
		tmpvar = variant(tmp_Ref_S, tmp_Ref_E, REF, ALT, _ReadPos)

		editd, editd_wo_intron, var_set, n_vars = calculate_edit_distance(tmpvar, Junctions, REFPOS)
		total_editd += editd
		total_n_vars += n_vars
		final_var_set.extend(var_set)

	return final_var_set, total_editd



def edit_distance_function_WA(var):

	mat = np.zeros((var.L_ref() + 1, var.L_alt() + 1)).astype(int)
	for i, char in enumerate(var.REF):
		mat[i + 1, 0] = (i + 1)*_Gap
	for j, char in enumerate(var.ALT):
		mat[0, j + 1] = (j + 1)*_Gap

	for i, char_ref in enumerate(var.REF):
		for j, char_rid in enumerate(var.ALT):
			subs = abs(_base_pair_score[(char_rid, char_ref)])
			mat[i + 1, j + 1] = min(mat[i, j] + subs, 
									mat[i, j + 1] + _Gap, 
									mat[i + 1, j] + _Gap)

	return mat[-1][-1]




def edit_distance_function_internal_aln(var, Junctions, REFPOS):

	def decompose_seq(seq, K, _junctions):

		intronic_seqs = {}
		for jxn_s, jxn_e in sorted(list(_junctions)):
			if var.g_start <= jxn_e and var.g_end >= jxn_s:
				pos = jxn_s - var.g_start
				intronic_seqs[jxn_s - var.g_start] = var.REF[pos : pos + (jxn_e - jxn_s)]
		ov_nodes = []

		j = 0
		while j < len(seq) - (K - 2):
			if j in intronic_seqs:
				intronic_seq = intronic_seqs[j]
				ov_nodes.append(intronic_seq)
				j = j + len(intronic_seq) - (K - 2)
			else:
				node = seq[j : j + (K - 1)]
				ov_nodes.append(node)
				j += 1
			if len(ov_nodes) > int(1e3):
				j = len(seq)
				ov_nodes = []

		return ov_nodes

	ALT_path = decompose_seq(var.ALT, _min_complex_length, [])
	REF_path = decompose_seq(var.REF, _min_complex_length, Junctions) 

	#### Too complex
	if len(REF_path)*len(ALT_path) > int(1e4) or len(REF_path)*len(ALT_path) == 0:
		return [var], edit_distance_lazy_calc(var)


	mat = np.zeros((len(ALT_path), len(REF_path))).astype(int)
	max_scores = {}

	for i in range(mat.shape[0]):
		cum_j = 0
		for j in range(mat.shape[1]):
			if ALT_path[i] == REF_path[j]:
				if i > 0 and j > 0:
					mat[i][j] = 1 + mat[i - 1][j - 1] 
				else:
					mat[i][j] = 1 

				mat_score = mat[i][j]
				
				if mat_score >= 2:
					origin = (i - mat_score + 1, cum_j - mat_score + 1)
					if origin not in max_scores or mat_score > max_scores[origin][0]:
						max_scores[origin] = (mat_score, i, cum_j)

			cum_j += len(REF_path[j]) - _min_complex_length + 2 


	del mat

	all_var_sets = []

	top = 1
	while max_scores and top <= 3:
		top += 1
		
		origin_max_score = max(max_scores, key = max_scores.get)

		RidS, RefS = origin_max_score

		(max_score, RidE, RefE) = max_scores[origin_max_score]
		max_scores.pop(origin_max_score)

		var_set = []

		### upstream of the aligned region
		Pfx, Sfx, S2, S1 = trim_sequences(var.ALT[:RidS], var.REF[:RefS])
		assert len(S1) + Pfx + Sfx == len(var.REF[:RefS]), "PP1 {} {} {}".format(S1, Pfx, Sfx, var.REF[:RefS])
		assert len(S2) + Pfx + Sfx == len(var.ALT[:RidS]), "PP2 {} {} {}".format(S2, Pfx, Sfx, var.ALT[:RidS])

		P1s, P1e = var.g_start + Pfx, var.g_start + RefS - Sfx
		P2s, P2e = Pfx, RidS - Sfx

		assert (P1e - P1s) == len(S1)
		assert (P2e - P2s) == len(S2)

		_ReadPos = {}
		for RI, (s, e) in var.RIPos.items():
			_ReadPos[RI] = (s + P2s, s + P2s + len(S2))
		_tmp_var = variant(P1s, P1e, S1, S2, _ReadPos)
		var_set.append(_tmp_var)

		
		### aligned region
		S1, S2 = var.REF[RefS : RefE], var.ALT[RidS : RidE]
		assert S1 == S2, "{} != {}".format(S1, S2)
		
		### downstream of the aligned region
		Pfx, Sfx, S2, S1 = trim_sequences(var.ALT[RidE:], var.REF[RefE:])
		assert len(S1) + Pfx + Sfx == len(var.REF[RefE:]), "PP3 {} {} {}".format(S1, Pfx, Sfx, var.REF[RefE:])
		assert len(S2) + Pfx + Sfx == len(var.ALT[RidE:]), "PP4 {} {} {}".format(S2, Pfx, Sfx, var.ALT[RidE:])

		P1s, P1e = var.g_start + RefE + Pfx, var.g_end - Sfx
		P2s, P2e = RidE + Pfx, len(var.ALT) - Sfx

		assert (P1e - P1s) == len(S1)
		assert (P2e - P2s) == len(S2)
		
		_ReadPos = {}
		for RI, (s, e) in var.RIPos.items():
			_ReadPos[RI] = (s + P2s, s + P2s + len(S2))
		_tmp_var = variant(P1s, P1e, S1, S2, _ReadPos)
		var_set.append(_tmp_var)
		all_var_sets.append(var_set)

	
	all_var_editd = {}
	for j, var_set in enumerate(all_var_sets):
		var_set_editd = 0
		var_set_editd_wo_intron = 0
		var_set_copy  = []
		
		for _tmp_var in var_set:
			EDITVAL = calculate_edit_distance(_tmp_var, Junctions, REFPOS)
			var_set_editd += EDITVAL[0]
			var_set_editd_wo_intron += EDITVAL[1]
			var_set_copy.extend(EDITVAL[2])

		all_var_sets[j]  = var_set_copy
		all_var_editd[j] = (var_set_editd_wo_intron, var_set_editd)

	try:
		min_j = min(all_var_editd, key = all_var_editd.get)
		Final_Var_Set   = all_var_sets[min_j]
		Final_Var_EditD = all_var_editd[min_j][1]
		
	except:
		Final_Var_Set   = [var]
		Final_Var_EditD = edit_distance_lazy_calc(var)

	return Final_Var_Set, Final_Var_EditD



def merge_introns(VAR_LIST, SM):
    
    intron_coords = HTSeq.GenomicArrayOfSets(chroms = "auto", stranded = False)

    sorted_introns = []
    
    for POS, ALLELE_dict in VAR_LIST.items():
        if 'INTRON' in POS:
            for ALLELE in ALLELE_dict:
            	if SM in ALLELE_dict[ALLELE]:
	                sorted_introns.append((POS, ALLELE))

    sorted_introns.sort(key = lambda x : x[0][1]) 

    sorted_introns_norep = []

    for intron in sorted_introns:
        POS, ALLELE = intron
        (CHROM, START, var_group), (REF, ALT, END) = POS, ALLELE

        if END - START < 20:
            continue  	  

        similar_found = False
        minD = 6 + len(ALT)

        for j in range(1, 6):
            if j > len(sorted_introns_norep): 
                break

            prev_POS, prev_ALLELE = sorted_introns_norep[-j]

            pSTART = prev_POS[1]
            pEND   = prev_ALLELE[2]

            if abs(START - pSTART) <= minD and abs(END - pEND) <= minD:
                similar_found = True
                VAR_LIST[prev_POS][prev_ALLELE][SM]["READS"].update(VAR_LIST[POS][ALLELE][SM]["READS"])
                iGI = HTSeq.GenomicInterval(CHROM, pSTART, pEND)
                intron = sorted_introns_norep[-j]
                break
        
        if not similar_found:
            sorted_introns_norep.append(intron)
            iGI = HTSeq.GenomicInterval(CHROM, START, END)
        
        intron_coords[iGI] += intron
        

    for iGI, intron_set in intron_coords.steps():

        if iGI.length > 20 and len(intron_set) >= 2:
            ip_id = (iGI.chrom, iGI.start, "INTRONIC_PART")
            ip_allele = ('', '', iGI.end)

            VAR_LIST[ip_id][ip_allele] = {SM : {"READS" : {}, "FEAT" : {}, "IP_ALLELES": [], "IP_ALLELES_AC" : []} }
            VAR_LIST[ip_id][ip_allele][SM]['FEAT']['FILTER'] = "PASS"
            VAR_LIST[ip_id][ip_allele][SM]['FEAT']['QUAL']   = 10.0
            VAR_LIST[ip_id][ip_allele]['variant_type'] = "INTRONIC_PART"
            VAR_LIST[ip_id][ip_allele]['variant_id']   = str(iGI)
            
            intron_set = list(intron_set)

            for i, intron in enumerate(intron_set):
                i_reads = VAR_LIST[intron[0]][intron[1]][SM]["READS"]

                VAR_LIST[ip_id][ip_allele][SM]["IP_ALLELES"].append(intron) 

                AC = 0
                for read_name, read_allele in i_reads.items():
                    if read_allele:
                        VAR_LIST[ip_id][ip_allele][SM]["READS"][read_name] = i
                        AC += 1

                VAR_LIST[ip_id][ip_allele][SM]["IP_ALLELES_AC"].append(AC)


    return VAR_LIST



def count_tandem_repeat_around_variant(var, REFPOS):

	tmp_vals = []
	for repLen in range(1, 5):
		for (l1, l2) in [(var.g_start, var.g_start + repLen), (var.g_start - repLen, var.g_start)]:

			repSeq = REFPOS.get_sequence(l1, l2)
			
			repCount_up = 0
			start_pos = l1
			while start_pos - repLen > REFPOS.gi.start + 2 and REFPOS.get_sequence(start_pos - repLen, start_pos) == repSeq:
				repCount_up += 1
				start_pos -= repLen

			repCount_dn = 0
			end_pos = l1
			while end_pos + repLen < REFPOS.gi.end - 2 and REFPOS.get_sequence(end_pos, end_pos + repLen) == repSeq:
				repCount_dn += 1
				end_pos += repLen

			repCount = repCount_up + repCount_dn
			tmp_vals.append((start_pos, end_pos, repSeq, repCount))

	(repeat_pos, repeat_end, repSeq, REP_COUNT) = max(tmp_vals, key = lambda x : x[3])

	REP_REGION = HTSeq.GenomicInterval(REFPOS.gi.chrom, repeat_pos - 1, repeat_end + 1)

	return REP_COUNT, REP_REGION



def trim_var_with_var(var1, var2):
	if var1 is None or var2 is None:
		return None
	elif var2.REF not in var1.REF or var2.ALT not in var1.ALT:
		return None
	elif var2.REF == var1.REF and var2.ALT == var1.ALT:
		return None
	elif var2.L_ref() > var1.L_ref() or var2.L_alt() > var1.L_alt():
		return None
	elif var2.get_variant_type() == "NO_VARIANT":
		return None

	if var2.g_start == var1.g_start:
		Pfx_ref, Sfx_ref, REF_12, ref_2c = trim_sequences(var1.REF, var2.REF, 'prefix')
		Pfx_alt, Sfx_alt, ALT_12, alt_2c = trim_sequences(var1.ALT, var2.ALT, 'prefix')

	elif var2.g_end == var1.g_end:
		Pfx_ref, Sfx_ref, REF_12, ref_2c = trim_sequences(var1.REF, var2.REF, 'sufix')
		Pfx_alt, Sfx_alt, ALT_12, alt_2c = trim_sequences(var1.ALT, var2.ALT, 'sufix')
	else:
		return None

	if ref_2c or alt_2c:
		return None

	_ReadPos = {}
	for r, (r_start, r_end) in var1.RIPos.items():
		_ReadPos[r] = (r_start + Pfx_alt, r_end - Sfx_alt)

	variant_3 = variant(var1.g_start + Pfx_ref, var1.g_end - Sfx_ref, REF_12, ALT_12, _ReadPos)
	variant_3.trim(min_match = 1)

	return variant_3




def find_equivalent_indels(VAR_COLLECTION, REFPOS):

	### First trim
	for i, variant in enumerate(VAR_COLLECTION):
		VAR_COLLECTION[i].trim(min_match = 1)
		if variant.get_variant_type() == "NO_VARIANT":
			VAR_COLLECTION[i] = None
		

	i = 0
	while i < len(VAR_COLLECTION):

		variant_1 = VAR_COLLECTION[i]
		
		if variant_1 is None or variant_1.get_variant_type() != "COMPLEX":
			i += 1
			continue
		
		split_candidates = {}
		for variant_2 in VAR_COLLECTION:
			variant_3 = trim_var_with_var(variant_1, variant_2)
			
			if variant_3 is None:
				continue
			
			e2 = calculate_edit_distance(variant_2, [], REFPOS)
			e3 = calculate_edit_distance(variant_3, [], REFPOS)
			split_candidates[(variant_2, variant_3)] = e2[0] + e3[0]
		

		if split_candidates:

			(variant_2, variant_3) = min(split_candidates, key = split_candidates.get)
			if split_candidates[(variant_2, variant_3)]	< 100.1:
				VAR_COLLECTION.append(variant_3)
				_ReadPos = {}
				Pfx = variant_2.g_start - variant_1.g_start
				Sfx = variant_2.g_end - variant_1.g_end
				for RI, (r_start, r_end) in variant_1.RIPos.items():
					_ReadPos[RI] = (r_start + Pfx, r_end - Sfx)
				variant_2.RIPos.update(_ReadPos)
				VAR_COLLECTION[i] = None
		i += 1
	

	### Find repeats around variants and move it to consensus location (beggining of rep)
	for i, variant in enumerate(VAR_COLLECTION):
		if variant is None:
			continue
		elif variant.get_variant_type() == "INTRON":
			variant.rep_region = HTSeq.GenomicInterval(REFPOS.gi.chrom, variant.g_start - 1, variant.g_end + 1)
			continue
		elif variant.get_variant_type() == "SNV":
			REP_COUNT, REP_REGION = count_tandem_repeat_around_variant(variant, REFPOS)
			variant.rep_region = HTSeq.GenomicInterval(REFPOS.gi.chrom, variant.g_start, variant.g_end)
			variant.rep_count  = REP_COUNT
			continue
		else:
			REP_COUNT, REP_REGION = count_tandem_repeat_around_variant(variant, REFPOS)
			variant.rep_count  = REP_COUNT
			variant.rep_region = REP_REGION

	### Find equivalent indels (same haplotype)
	for i, variant_1 in enumerate(VAR_COLLECTION):
		if variant_1 is None:
			continue
		if variant_1.get_variant_type() in ("INSERTION", "DELETION"):
			variant_1.leftalign(REFPOS)
		for j, variant_2 in enumerate(VAR_COLLECTION):
			if i == j or variant_2 is None:
				continue
			if variant_1.same_variant(variant_2, REFPOS):
				if variant_2.get_variant_type() in ("INSERTION", "DELETION"):
					variant_2.leftalign(REFPOS)
				VAR_COLLECTION[i].RIPos.update(variant_2.RIPos)
				VAR_COLLECTION[j] = None

	### Revome variants outside of RC coordinates
	for i, variant in enumerate(VAR_COLLECTION):
		if variant is None:
			continue
		elif variant.get_variant_type() == "INTRON":
			continue
		else:
			if REFPOS.gi.start <= variant.g_start and variant.g_start <= REFPOS.gi.end:
				pass
			else:
				VAR_COLLECTION[i] = None

	### Final trimming
	for i, variant in enumerate(VAR_COLLECTION):
		if variant is None:
			continue
		elif variant.get_variant_type() == "NO_VARIANT":
			VAR_COLLECTION[i] = None
			continue
		VAR_COLLECTION[i].trim(min_match = 1)
		if variant.get_variant_type() in ("INSERTION", "DELETION"):
			VAR_COLLECTION[i].leftalign(REFPOS)
	
	return [v for v in VAR_COLLECTION if v is not None]




#!/usr/bin/python
import os
import sys
import time 
import copy 

try:
	import psutil
	def print_time_stamp(Message = ""):
		process = psutil.Process()
		Memory  = process.memory_info()[0]/float(1024**3)
		timeS   = time.strftime("[%a, %d %b %Y, %H:%M:%S]")
		sys.stderr.write("{}\t{:.3f}\t{}\n".format(timeS, Memory, Message))  
		sys.stdout.flush() 
		sys.stderr.flush() 

except:
	def print_time_stamp(Message = ""):
		timeS = time.strftime("[%a, %d %b %Y %H:%M:%S]")
		sys.stderr.write("{}\t{}\n".format(timeS, Message))


def delete_reads(d1):
	for key, val in list(d1.items()):
		if key == "READS":
			d1['READS'].clear()
			
		elif isinstance(val, dict): 
			delete_reads(val)



def merge_copy(d1, d2):
    for key in d2:
        if key in d1 and isinstance(d1[key], dict) and isinstance(d2[key], dict):
            merge_copy(d1[key], d2[key])
        elif key in d1 and isinstance(d1[key], set) and isinstance(d2[key], set):
            d1[key] |= d2[key]
        else:
            d1[key] = copy.deepcopy(d2[key])

            
def rm_nested_dict(my_dict):
	my_dict.clear()
	
	# for key, sub_dict in list(my_dict.items()):
	# 	try:
	# 		rm_nested_dict(sub_dict)
	# 	except:
	# 		pass
	# 	my_dict.pop(key)


def rm_nested_list(my_list):
	while my_list:
		del my_list[0]


def get_size(obj, seen=None):
	"""Recursively finds size of objects"""

	size = sys.getsizeof(obj)
	if seen is None:
		seen = set()
	obj_id = id(obj)
	if obj_id in seen:
		return 0

	seen.add(obj_id)
	if isinstance(obj, dict):
		size += sum([get_size(v, seen) for v in obj.values()])
		size += sum([get_size(k, seen) for k in obj.keys()])
	elif hasattr(obj, '__dict__'):
		size += get_size(obj.__dict__, seen)
	elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
		size += sum([get_size(i, seen) for i in obj])

	return size













#!/usr/bin/python
import time
import sys
import gc
import os
import re
import copy
import HTSeq
import collections
import multiprocessing 
import _pickle as pickle
import numpy as np
from optparse import OptionParser, OptionGroup

import importlib.resources as pkg_resources

from . import glm
from . import gqv_utils
from . import gqv_dbg_utils
from . import gqv_bam_utils
from . import gqv_glm
from . import gqv_vartool
from . import gqv_software_management as SOFTMAN
from . import gqv_mutual_information  as MUTINFO




def process_read_cluster_1(ARGs):
	
	RC_info, options, bam_file, SM = ARGs

	RC_Index, chrom, strand, RC_Start, RC_End, MaxCov = RC_info 

	genome_faidx = gqv_utils.read_genome_fasta(options.Genome)

	mp_id = multiprocessing.current_process().name
	
	### Obtain reads that overlap RC
	RC_vals = gqv_bam_utils.get_RC_reads(bam_file, RC_info, genome_faidx, SM, options)

	if RC_vals is None:
		return None
	else:
		SetOfReadNames, REFPOS, SetOfReadGroups = RC_vals
	
	### Construct de Bruijn graph
	db = gqv_dbg_utils.construct_de_Bruijn_graph(SetOfReadGroups, options) 
	
	### Process de Bruijn graph
	xdb, SetOfReadIndexes = gqv_dbg_utils.Process_de_Bruijn_Graph(SetOfReadGroups, options, db) 

	del db

	REF_INDEX = SetOfReadGroups[0]['Index'] 


	### DFT algorithm for src-snk pairs compressed
	Src_Snk_pairs = gqv_dbg_utils.find_Source_Target_pairs(xdb, REF_INDEX, SetOfReadIndexes, options)  
	del xdb
	
	if not Src_Snk_pairs:
		return None

	### Find variants for every read index
	SetOfVariants = gqv_utils.assign_vars_to_reads(Src_Snk_pairs, SetOfReadIndexes, 
													SetOfReadGroups, options, REFPOS) 
	
	SOFTMAN.rm_nested_dict(Src_Snk_pairs)
	
	### Process, filter and modify variants
	SetOfVariants = gqv_vartool.find_equivalent_indels(SetOfVariants, REFPOS) 

	### Obtian all overlapping read indexes for every variant
	SetOfVariants = gqv_utils.overlap_vars_and_reads(SetOfVariants, SetOfReadIndexes, 
										   SetOfReadGroups, REFPOS)
	SOFTMAN.rm_nested_dict(SetOfReadGroups)
	SOFTMAN.rm_nested_list(SetOfReadIndexes) 

	### Feature collection for every variant
	VAR_LIST = gqv_utils.feature_collection(SetOfVariants, SetOfReadNames, options, SM, REFPOS)

	### Genotyping table
	VAR_LIST = gqv_utils.table_genotyping(VAR_LIST, SM, options)	
	
	return bam_file, SM, RC_info, SetOfVariants, VAR_LIST




def process_read_cluster_2(ARGs):

	RC_info, options, bam_file, SM, SetOfVariants, AllOfVariants = ARGs

	RC_Index, chrom, strand, RC_Start, RC_End, MaxCov = RC_info 

	# ----- get variants -------
	RefVariants = []
	RC_GI = HTSeq.GenomicInterval(chrom, RC_Start, RC_End)
	for v_gi, variant_set in AllOfVariants[RC_GI].steps():
		for variant in variant_set:
			if variant not in SetOfVariants:
				RefVariants.append(variant)

	# --------------------------------------------

	if not RefVariants:
		return {}

	genome_faidx = gqv_utils.read_genome_fasta(options.Genome)

	mp_id = multiprocessing.current_process().name

	### Obtain reads that overlap RC
	RC_vals = gqv_bam_utils.get_RC_reads(bam_file, RC_info, genome_faidx, SM, options)

	if RC_vals is None:
		return {}
	else:
		SetOfReadNames, REFPOS, SetOfReadGroups = RC_vals

	### Obtian all overlapping read indexes for every variant
	SetOfVariants = gqv_utils.overlap_vars_and_reads_ref(RefVariants, SetOfReadGroups)

	SOFTMAN.rm_nested_dict(SetOfReadGroups)

	### Feature collection for every variant
	VAR_LIST = gqv_utils.feature_collection_ref(SetOfVariants, SetOfReadNames, options, SM, REFPOS)	

	return VAR_LIST




def parse_chroms(bam_files, options):
	region = options.search_region

	genome_chroms = gqv_utils.read_genome_fasta(options.Genome)

	bam_chroms = set()
	for bam_file in bam_files:
		bam_chroms |= set(gqv_bam_utils.get_chroms_from_bam(bam_file))

	common_chroms = set(genome_chroms.index) & bam_chroms
	
	Region_List = []
	
	r = re.match("([\d|\w]+):(\d+)\-(\d+)", region)

	if not common_chroms:
		raise ValueError("Invalid genome reference\n")

	elif region == "All":
		# all super-clusters
		Region_List = gqv_bam_utils.find_super_read_clusters(bam_files, options) 
	
	elif bool(r):
		chrom, fetch_start, fetch_end = r.groups()

		# all super-clusters that overlap this region
		if chrom not in common_chroms:
			raise ValueError(f"Region is not valid {region}\n")

		Region_List = gqv_bam_utils.find_super_read_clusters(bam_files, options, 
																chrom   = chrom, 
																f_start = int(fetch_start), 
																f_end   = int(fetch_end)) 
	
	else:
		# all super-clusters in this chrom
		if region in common_chroms: 		
			Region_List = gqv_bam_utils.find_super_read_clusters(bam_files, options, 
																	chrom = region)
		else:
			raise ValueError(f"Region is not valid {region}\n")

	return Region_List




def main():
	args = sys.argv

	description = """A variant caller and variant analysis tool for scRNA-seq data."""

	usage = """\n\tscAllele -b <file.bam> -g <genome.fa> -o <output prefix>"""

	strandedness_options = ['fr-firststrand', 'fr-secondstrand', 'fr-unstrand']


	parser = OptionParser(usage = usage, description = description)
	
	parser.add_option("-b", "--input-bam",   
		dest = "input_bam", 
		help = "[Required] Input bam file, (or comma-seprated list of bam files) sorted and indexed")
	parser.add_option("-o", "--output-vcf",  
		dest = "output_prefix", 
		help = "[Required] Prefix of the output files")
	parser.add_option("-g", "--genome-file", 
		dest = "Genome", 
		help = "[Required] Reference genome file (fasta format)")


	filtering_options = OptionGroup(parser, "Filtering Options")

	filtering_options.add_option("--AB", 
		dest = "minRatioVar", 
		type = "float",  
		default = 1e-2, 
		help = "Minimum allelic ratio for the variant allele. Default: 0.01")
	filtering_options.add_option("--AC", 
		dest = "minCountVar", 
		type = "int",    
		default = 2,    
		help = "Minimum read depth supporting the variant allele. Default: 2")
	filtering_options.add_option("--DP", 
		dest = "minCoverage", 
		type = "int",    
		default = 5,   
		help = "Minimum read depth at the position of the variant. Default: 5")
	filtering_options.add_option("--min-base_position", 
		dest = "minReadPos", 
		type = "int", 
		default = 10,
		help = "Minimum mean distance for the variant from the read ends. Default = 7")
	filtering_options.add_option("--min-base_quality", 
		dest = "minBaseQual", 
		type = "int", 
		default = 20,
		help = "Minimum mean base quality score to consider SNPs. Default = 20")

	parser.add_option_group(filtering_options)

	
	run_mode_options = OptionGroup(parser, "Run Mode Options")
	
	run_mode_options.add_option("--run_mode",  
		dest = "run_mode", 
		default = 'Full', 
		help = "Select <Variant_Caller> <Full> or <Training> mode. Default: Full")
	run_mode_options.add_option("--glm_clf_name",  
		dest = "glm_classifier_name", 
		default = None, 
		help = "Prefix of the GLM pickle objects with the GLM models")
	
	parser.add_option_group(run_mode_options)


	linkage_options = OptionGroup(parser, "Linkage Options")
	
	linkage_options.add_option("--link_min_count",  dest = "link_min_count", 
		default =  10,
		type = "int", 
		help = "Minimum number of common reads for linkage analysis. Default = 10")
	linkage_options.add_option("--link_min_mi",  dest = "link_min_mi", 
		default = 0.52, 
		type = "float",
		help = "Minimum mutual information for linkage analysis. Default = 0.52")
	
	parser.add_option_group(linkage_options)


	advanced_options = OptionGroup(parser, "Advanced Options")

	advanced_options.add_option("-c", "--region", 
		dest = "search_region",     
		default = 'All', 
		help = "Limit search to this region (chrom:start-end or chrom). Default: All")
	advanced_options.add_option("-n", "--nodes", 
		dest = "nodes", 
		type = "int", 
		default = 64, 
		help = "Number of threads for parallel execution. Default = 64")
	advanced_options.add_option("--min-map_quality", 
		dest = "minMapQ", 
		type = "int", 
		default = 40,
		help = "Minimum mapping quality to keep a read. Default = 40")
	advanced_options.add_option("--max_soft_clip", 
		dest = "maxSoftClip", 
		type = "int", 
		default = 5,
		help = "Maximum length of soft-clipping allow at each end of the read. Default = 5")
	advanced_options.add_option("--kmer-size", 
		dest = "kmer", 
		type = "int", 
		default = 15,
		help = "k-mer size for the de-Bruijn graph assembly. Default: 15")
	advanced_options.add_option("--strandedness", 
		dest = "strandedness", 
		default = "fr-unstrand",
		help = "Select from {}. Default: 'fr-unstrand'".format(strandedness_options))
	advanced_options.add_option("--maxSeqErrorRate", 
		dest = "maxSeqErrorRate", 
		type = "float", 
		default = 1e-2,
		help = "Maximum estimate of sequencing error rate. Default: 0.01")
	advanced_options.add_option("--Ploidy", 
		dest = "ploidy", 
		type = "int", 
		default = 2,
		help = "Maximum ploidy to be considered. Default: 2.")

	parser.add_option_group(advanced_options)

	(options, Args) = parser.parse_args()

	if len(sys.argv[1:]) < 3:
		parser.print_help()
		sys.exit(1)


	if not os.path.exists(options.Genome):
		parser.error('Genome file does not exist')

	if not options.output_prefix:   
		parser.error('Output file prefix not given')
	
	if options.strandedness not in strandedness_options:
		parser.error("\nStrandedness option: {} not valid. \
			Use the -h option to view usage\n".format(strandedness))

	if options.kmer > 20 or options.kmer < 10:
		SOFTMAN.print_time_stamp("WARNING: We recomend a k-mer size of 10-20")

	if options.run_mode in ("Full", "Variant_Caller"):
		if options.glm_classifier_name is not None:
			f_d = open(options.glm_classifier_name + '.DELETION.glm.pickle', 'rb')
			f_i = open(options.glm_classifier_name + '.INSERTION.glm.pickle', 'rb')
			f_s = open(options.glm_classifier_name + '.SNP.glm.pickle', 'rb')
		else:
			f_d = pkg_resources.open_binary(glm, 'GM12878_smartseq2.all.feature_matrix.tab.DELETION.glm.pickle')
			f_i = pkg_resources.open_binary(glm, 'GM12878_smartseq2.all.feature_matrix.tab.INSERTION.glm.pickle')
			f_s = pkg_resources.open_binary(glm, 'GM12878_smartseq2.all.feature_matrix.tab.SNP.glm.pickle')

		GLM_DEL = pickle.load(f_d)
		GLM_INS = pickle.load(f_i)
		GLM_SNP = pickle.load(f_s)

	if options.run_mode not in ("Variant_Caller", "Full", "Training"):
		parser.error("Run mode {} not valid\n".format(options.run_mode))


	

	sys.setrecursionlimit(int(1e5))

	SOFTMAN.print_time_stamp("COMMAND = '{}'".format(' '.join(args)))
	
	SOFTMAN.print_time_stamp("Running {} on {} mode".format(args[0], options.run_mode))


	bam_list = gqv_bam_utils.read_bam_input(options.input_bam)

	Search_Regions = parse_chroms(bam_list, options)

	SAMPLE_list = [(bam_file, gqv_bam_utils.get_sample_name(bam_file)) for bam_file in bam_list]


	if options.run_mode in ("Full", "Variant_Caller"):
		gqv_utils.write_vcf_header(Search_Regions, options.Genome, 
									options.output_prefix + ".vcf", SAMPLE_list)
		gqv_utils.write_introns_header(options.output_prefix + ".intronic_parts.bed")

	if options.run_mode in ("Full"):
		gqv_utils.write_mutinfo_header(options.output_prefix + ".mi_summary.tab")


	n = min(options.nodes, multiprocessing.cpu_count())
	pool = multiprocessing.Pool(n)


	for Search_Region in Search_Regions:	

		(chrom, fetch_start, fetch_end) = Search_Region
		L = fetch_end - fetch_start
		SOFTMAN.print_time_stamp("Processing contig {0[0]}:{0[1]}-{0[2]} Length = {1}".format(Search_Region, L))
			
		ALL_READ_CLUSTERS = collections.defaultdict(dict)
		chrom_VAR_LIST    = collections.defaultdict(dict)
		chrom_VAR_MUTINFO = []
		
		ARGs = []
		for (bam_file, SM) in SAMPLE_list:	
			RCs = gqv_bam_utils.find_read_clusters(bam_file, options, chrom, fetch_start, fetch_end)
			ARGs += [(rc, options, bam_file, SM) for rc in RCs]

		
		### ----- Pool processing -----

		raw_VAR_LIST = []

		for rc_output in pool.imap_unordered(process_read_cluster_1, ARGs): 
			if rc_output is None: 
				continue
			
			bam_file, SM, rc, SetOfVariants, rc_VAR_LIST = rc_output
			raw_VAR_LIST.extend(copy.deepcopy(SetOfVariants))
			SOFTMAN.merge_copy(chrom_VAR_LIST, rc_VAR_LIST)
			ALL_READ_CLUSTERS[(SM, bam_file)][tuple(rc)] = SetOfVariants				

		gc.collect()

		
		for (bam_file, SM) in SAMPLE_list:

			if options.run_mode in ("Full", "Variant_Caller"):
				chrom_VAR_LIST = gqv_vartool.merge_introns(chrom_VAR_LIST, SM)
				gqv_glm.update_vars(GLM_INS, GLM_DEL, GLM_SNP, chrom_VAR_LIST, SM, options)

			if options.run_mode in ("Full", ):
				mi_outlines = MUTINFO.mi_parse_variants(chrom_VAR_LIST, SM, options)
				chrom_VAR_MUTINFO.extend(mi_outlines)

		raw_VAR_LIST = gqv_utils.process_var_list(chrom, raw_VAR_LIST, chrom_VAR_LIST)

		
		ARGs = []
		for (SM, bam_file), RCs in ALL_READ_CLUSTERS.items():

			if not RCs: 
				continue
			
			for rc, SetOfVariants in RCs.items():
				ARGs.append((rc, options, bam_file, SM, SetOfVariants, raw_VAR_LIST))

		for rc_VAR_LIST in pool.imap_unordered(process_read_cluster_2, ARGs): 
			SOFTMAN.merge_copy(chrom_VAR_LIST, rc_VAR_LIST)


		
		if options.run_mode in ("Full", "Variant_Caller"):

			gqv_utils.write_introns_bed(chrom_VAR_LIST, options.output_prefix + ".intronic_parts.bed")
			chrom_VAR_LIST = gqv_utils.merge_SM_VAR_LIST(chrom_VAR_LIST)
			chrom_VAR_LIST = gqv_vartool.merge_introns(chrom_VAR_LIST, 'merged_SM')
			gqv_glm.update_vars(GLM_INS, GLM_DEL, GLM_SNP, chrom_VAR_LIST, 'merged_SM', options)

		if options.run_mode in ("Full",):

			mi_outlines = MUTINFO.mi_parse_variants(chrom_VAR_LIST, 'merged_SM', options)
			chrom_VAR_MUTINFO.extend(mi_outlines)


		SOFTMAN.delete_reads(chrom_VAR_LIST)


		if options.run_mode in ("Full", "Variant_Caller"):
			gqv_utils.write_vcf_file(chrom_VAR_LIST, options.output_prefix + ".vcf", SAMPLE_list)
			gqv_utils.write_introns_bed(chrom_VAR_LIST, options.output_prefix + ".intronic_parts.bed")

		if options.run_mode in ("Full",):
			gqv_utils.write_mutinfo_file(chrom_VAR_MUTINFO, options.output_prefix + ".mi_summary.tab")

		if options.run_mode == "Training":
			gqv_utils.write_feat_file(ALL_VARS_LIST, options.output_prefix + ".feature_matrix.tab")


		gqv_utils.write_readcluster_file(ALL_READ_CLUSTERS, 
									options.output_prefix + ".read_cluster_info.tab")


		SOFTMAN.rm_nested_dict(chrom_VAR_LIST)
		SOFTMAN.rm_nested_dict(ALL_READ_CLUSTERS)
		gc.collect()


	pool.close()
	pool.terminate()
	SOFTMAN.print_time_stamp("DONE!")



if __name__ == "__main__":
	main()








#!/usr/bin/python
import networkx as nx
import numpy as np ; np.seterr(all = "ignore")
import itertools
import copy
from collections import defaultdict, Counter
from time import time
from . import gqv_software_management as SOFTMAN


def construct_de_Bruijn_graph(Unique_Read_Seqs, options):
	G = nx.DiGraph()
	k = options.kmer
	for seq_id, attr in Unique_Read_Seqs.items():
		ReadSeq = attr['seq']
		n_i = len(attr['reads'])
		for a_j in range(len(ReadSeq) - k + 1):
			kmer_or = ReadSeq[a_j : a_j + k - 1]
			kmer_dt = ReadSeq[a_j + 1 : a_j + k]
			rname_node = (seq_id, a_j)
			try:
				G[kmer_or][kmer_dt]['n'] += n_i
				G[kmer_or][kmer_dt]['reads'].append(rname_node)
			except:
				G.add_edge(kmer_or, kmer_dt, reads = [rname_node], n = n_i)
	return G



def Process_de_Bruijn_Graph(SetOfReadGroups, options, db):
	k = options.kmer

	node_First  = SetOfReadGroups[0]['seq'][:k - 1]
	node_Last   = SetOfReadGroups[0]['seq'][-(k - 1):]
	node_Second = db.successors(node_First).__next__()

	xdb = nx.DiGraph()

	dbg_compress_graph(db, xdb, k, node_First, node_Second)

	min_Allele_Ratio = min(options.maxSeqErrorRate * 2, 0.05)
	dbg_remove_lowly_covered_edges(xdb, 0, min_Allele_Ratio)

	all_nodes = list(xdb.nodes)[:]
	dbg_remove_alternative_ends(all_nodes, xdb, node_First, node_Last)


	Graph_Read_Tracker = defaultdict(dict)

	dbg_get_read_paths(xdb, Graph_Read_Tracker, k)

	SetOfReadIndexes = compress_read_paths(xdb, k, Graph_Read_Tracker, SetOfReadGroups)

	SOFTMAN.rm_nested_dict(Graph_Read_Tracker)

	return (xdb, SetOfReadIndexes)


def dbg_remove_lowly_covered_edges(db, REF_Index, min_Allele_Ratio):
    edges_to_remove = set()
    for node in db.nodes:
        OutNodesCountTotal = 0.0
        for node_out in db.successors(node):
            OutNodesCountTotal += db[node][node_out]['n']
        for node_out in db.successors(node):
            OutNodesCount  = db[node][node_out]['n']
            EdgeCountRatio = OutNodesCount/OutNodesCountTotal
            if not any(r[0] == REF_Index for r in db[node][node_out]['reads']):
                if EdgeCountRatio <= min_Allele_Ratio: # or OutNodesCount <= 1:
                    edges_to_remove.add((node, node_out))

        InNodesCountTotal = 0.0
        for node_in in db.predecessors(node):
            InNodesCountTotal += db[node_in][node]['n']
        for node_in in db.predecessors(node):
            InNodesCount  = db[node_in][node]['n']
            EdgeCountRatio = InNodesCount/InNodesCountTotal
            if not any(r[0] == REF_Index for r in db[node_in][node]['reads']):
                if EdgeCountRatio <= min_Allele_Ratio: # or InNodesCount <= 1:
                    edges_to_remove.add((node_in, node))

    for (a, b) in edges_to_remove:
        db.remove_edge(a, b)

    del edges_to_remove


def dbg_remove_alternative_ends(all_nodes, db, node_First, node_Last):	
	for node in all_nodes:
		next_nodes = list()
		if db.out_degree(node) == 0 and db.in_degree(node) == 0:
			db.remove_node(node)
		elif db.out_degree(node) == 1 and node in db.successors(node):
			db.remove_node(node)
		elif db.in_degree(node) == 1 and node in db.predecessors(node):
			db.remove_node(node)
		elif db.in_degree(node) == 0 and node != node_First:
			if db.out_degree(node) == 0:
				next_nodes = []
			else:
				next_nodes = db.successors(node)
			db.remove_node(node)
			dbg_remove_alternative_ends(next_nodes, db, node_First, node_Last)
		elif db.out_degree(node) == 0 and node != node_Last:
			if db.in_degree(node) == 0:
				next_nodes = []
			else:
				next_nodes = db.predecessors(node)
			db.remove_node(node)
			dbg_remove_alternative_ends(next_nodes, db, node_First, node_Last)
	


def dbg_compress_graph(db, xdb, k, prev_node, node):
	node_seq = list()
	OV = k - 1
	while db.out_degree(node) == 1 and db.in_degree(node) == 1 and "I" not in node:
		node_seq.append(node)
		node = db.successors(node).__next__()

	if not node_seq:
		curr_node = node
	else:
		curr_node = ''.join(n[0] for n in node_seq) + node_seq[-1][1:]

	if (prev_node, curr_node) not in xdb.edges:
			
		n_db = db[prev_node[-OV:]][curr_node[:OV]]['n']
		r_db = db[prev_node[-OV:]][curr_node[:OV]]['reads']
		xdb.add_edge(prev_node, curr_node, reads = r_db, ris = set(), n = n_db) 
				
		if node != curr_node[-OV:]:
			n_db = db[curr_node[-OV:]][node[:OV]]['n']
			r_db = db[curr_node[-OV:]][node[:OV]]['reads']
			xdb.add_edge(curr_node, node, reads = r_db, ris = set(), n = n_db) 

		for out_node in db.successors(node):
			dbg_compress_graph(db, xdb, k, node, out_node)



def dbg_get_read_paths(xdb, RP, k):
	for (cn, nn) in xdb.edges:			
		for (rname, kmer_index) in xdb[cn][nn]['reads']:
			index_offset = len(cn) - (k - 1)
			try:
				RP[rname][cn].append(kmer_index - index_offset)
			except:
				RP[rname][cn] = [kmer_index - index_offset]
			try:
				RP[rname][nn].append(kmer_index + 1)
			except:
				RP[rname][nn] = [kmer_index + 1]



def compress_read_paths(xdb, k, RP, SetOfReadGroups):
	OV = k - 2

	tmp1_SetOfReadIndexes = defaultdict(set)
	tmp2_SetOfReadIndexes = defaultdict(set)
	
	for RG, attr in sorted(SetOfReadGroups.items()):

		RG_Seq = attr['seq']
		if RG in RP:
			path_dict = RP.pop(RG)
		else:
			SetOfReadGroups.pop(RG)
			continue

		known_nodes = set()
		for LongNode, index_list in path_dict.items():
			for r_pos in index_list:
				known_nodes.add((r_pos, LongNode))

		known_nodes = list(known_nodes)
		known_nodes.sort(key = lambda j: j[0])

		first_node = known_nodes[0][1]
		cont_read_paths = {}
		cont_read_path_length = len(first_node)
		j_0 = 0
		j_i = 0
		for j_i in range(len(known_nodes) - 1):
			(curr_i, curr_node), (next_i, next_node) = known_nodes[j_i:j_i + 2]
			if curr_node in xdb.nodes and next_node in xdb.successors(curr_node):
				cont_read_path_length += len(next_node) - OV
			else:
				cont_read_paths[cont_read_path_length] = (j_0, j_i + 1)
				j_0 = j_i + 1
				cont_read_path_length = len(next_node)
		cont_read_paths[cont_read_path_length] = (j_0, j_i + 2)

		for i, (pathLen, (Max_s, Max_e)) in enumerate(cont_read_paths.items()): 
			convert_read_pos, PATH = map(list, zip(*known_nodes[Max_s : Max_e]))
			if RG == 0:
				LongSeq = PATH[0] + ''.join(p[OV:] for p in PATH[1:])
				assert LongSeq == RG_Seq, "{}\n!=\n{}".format(LongSeq, RG_Seq)
				REF_SEQ_PATH = np.array(PATH)
				REF_SEQ_PATH_str = ";".join(PATH)
				newRG = 0
			elif len(set(PATH)) == 1:
				continue
			else:
				newRG = "{}_{}".format(RG, i)
			
			SetOfReadGroups[newRG] = copy.deepcopy(SetOfReadGroups[RG])
			SetOfReadGroups[newRG]['Convert_Pos'] = convert_read_pos
			tmp1_SetOfReadIndexes[';'.join(PATH)].add(newRG)

		if RG != 0:
			SetOfReadGroups.pop(RG)

	for PATH_str in sorted(tmp1_SetOfReadIndexes.keys(), key = lambda x : len(x), reverse = True):

		RGs = tmp1_SetOfReadIndexes[PATH_str]

		PATH = PATH_str.split(";") 

		
		while xdb.in_degree(PATH[0]) == 1 and xdb.predecessors(PATH[0]).__next__() not in PATH \
			and not any(char in xdb.predecessors(PATH[0]).__next__() for char in ["I", "X", "Y"]):
			
			pred = xdb.predecessors(PATH[0]).__next__()
			PATH.insert(0, pred)
			for RG in RGs:
				SetOfReadGroups[RG]['Convert_Pos'].insert(0, np.nan)

		while xdb.out_degree(PATH[-1]) == 1 and xdb.successors(PATH[-1]).__next__() not in PATH \
			and not any(char not in xdb.successors(PATH[-1]).__next__() for char in ["I", "X", "Y"]):
			
			succ = xdb.successors(PATH[-1]).__next__()
			PATH.append(succ)
			for RG in RGs:
				SetOfReadGroups[RG]['Convert_Pos'].append(np.nan)
				

		PATH_str = ';'.join(PATH)
		tmp2_SetOfReadIndexes[PATH_str] |= RGs


	SetOfReadIndexes = []
	for Read_Index, (PATH_str, Read_Group_Set) in enumerate(tmp2_SetOfReadIndexes.items()):
		PATH = PATH_str.split(";") 
		INDEX = [0]
		icm = 0
		LongSeq = PATH[0]
		for j in range(len(PATH) - 1):
			LongSeq += PATH[j + 1][OV:]
			icm += len(PATH[j]) - OV
			INDEX.append(icm)

		read_obj = Read_Index_Object(INDEX, k, LongSeq, Read_Group_Set)

		for RG in Read_Group_Set:
			SetOfReadGroups[RG]['Index'] = Read_Index
			read_obj.n += len(SetOfReadGroups[RG]['reads'])
			blocks = SetOfReadGroups[RG]['blocks']
			offset = np.array(INDEX) - np.array(SetOfReadGroups[RG]['Convert_Pos'])
			offset = offset[~np.isnan(offset)][0]

			SetOfReadGroups[RG]['Convert_Pos'] = list(map(int, np.array(INDEX) - offset))

			for b_i in range(len(blocks) - 1):
				read_obj.junctions.add((blocks[b_i][1], blocks[b_i + 1][0]))

		for j in range(len(PATH) - 1):
			xdb[PATH[j]][PATH[j + 1]]['ris'].add(Read_Index)

		SetOfReadIndexes.append(read_obj)

	del tmp1_SetOfReadIndexes
	del tmp2_SetOfReadIndexes

	return SetOfReadIndexes



class Read_Index_Object:

    def __init__(self, path_index, k, long_seq, Read_Group_Set):
        self.PathIndex    = path_index
        self.PathIndexExt = path_index + [len(long_seq)]
        self.LongSeq = long_seq
        self.reads = Read_Group_Set
        self.n = 0
        self.k = k
        self.gStart = None
        self.gEnd = None
        self.junctions = set([])
        self.Events = defaultdict(set)
        self.EventsCount = 0
        self.mapping = []

    def PathSeq(self, return_index = False):
        array = []
        i2 = 0
        for i in range(len(self.PathIndex) - 1):
            i1, i2 = self.PathIndex[i : i + 2]
            node = self.LongSeq[i1 : i2 + self.k - 2]
            array.append(node)

        array.append(self.LongSeq[i2:])
        if return_index:
            node_index = defaultdict(list)
            for i, node in enumerate(array):
                node_index[node].append(i)
            return (array, node_index)
        else:
            return np.array(array)

    def add_mapping(self, mapping_list):
        if mapping_list:
            self.mapping = mapping_list
            self.gStart = mapping_list[0][2]
            self.gEnd = mapping_list[-1][3]

    def add_event(self, Ref_Coords, Rid_Coords, f_Seq, d_Seq):
        if len(d_Seq) - len(f_Seq) <= 20 and (d_Seq or f_Seq):
            try:
                self.Events[Rid_Coords].add(Ref_Coords)
            except:
                self.Events[Rid_Coords] = set([Ref_Coords])
            self.EventsCount += 1

    def find_position(self, POS, END):
        rid_pos, rid_end = (None, None)
        for i, (rid_s, rid_e, ref_s, ref_e, tag) in enumerate(self.mapping):
            if tag == 'REF':
                if ref_s <= POS and POS < ref_e and rid_pos is None:
                    rid_pos = (rid_s + POS - ref_s, i)
                if ref_s < END and END <= ref_e and rid_end is None:
                    rid_end = (rid_s + END - ref_s, i)
            elif tag == 'ALT1' and (ref_e - ref_s) > 20:
                if ref_s <  POS and END < ref_e:
                    rid_pos, rid_end = (None, None)
            elif tag == 'ALT1':
                if ref_s <= POS and POS <= ref_e:
                    rid_pos = (rid_s, i)
                if ref_s <= END and END <= ref_e:
                    rid_end = (rid_e, i)

        if rid_pos is None or rid_end is None:
            if POS == END and rid_pos is not None:
                j = rid_pos[1]
                return (rid_pos[0], rid_pos[0], self.mapping[j][4])
            elif POS == END and rid_end is not None:
                j = rid_end[1]
                return (rid_end[0], rid_end[0], self.mapping[j][4])
            else:
                return (None, None, None)
        else:
            if rid_pos[1] == rid_end[1]:
                j = rid_pos[1]
                return (rid_pos[0], rid_end[0], self.mapping[j][4])
            else:
                return (rid_pos[0], rid_end[0], 'ALT1')

    def delete_events(self):
        SOFTMAN.rm_nested_dict(self.Events)



def is_contained(path_1, path_2):
    i = -1
    overlap0 = False
    while path_2[0] in path_1[i + 1: ]:
        i = path_1.index(path_2[0], i + 1)
        n = len(path_2)
        if path_2 == path_1[i : i + n]:
            overlap0 = True 
            break
    return i, overlap0





def find_Source_Target_pairs(xdb, REF_Index, SetOfReadIndexes, options):
	OV = options.kmer - 2
	REFERENCE_PATH = SetOfReadIndexes[REF_Index].PathSeq()

	src = defaultdict(list) 
	snk = defaultdict(list)  
	SRCSNK_pairs = defaultdict(dict)
	bubbles_from_src = defaultdict(set)
	
	for i, node in enumerate(REFERENCE_PATH):
		if xdb.out_degree(node) > 1:
			src[node].append(i) 
		if xdb.in_degree(node) > 1:
			snk[node].append(i) 

	for Read_Index, r in enumerate(SetOfReadIndexes):
		if Read_Index == REF_Index:
			continue
		rid_path = r.PathSeq()

		bubble_dict = defaultdict(list)
		for j, node in enumerate(rid_path):
			if node in src:
				bubble_dict["src"].append(j)
			if node in snk:
				bubble_dict["snk"].append(j)

		for src_i, snk_i in itertools.product(bubble_dict["src"], bubble_dict["snk"]):
			snk_seq, src_seq = rid_path[snk_i], rid_path[src_i]
			if snk_i >= src_i and max(snk[snk_seq]) >= min(src[src_seq]):
				SrcSnk_Seq = ''.join(x[OV:] for x in rid_path[src_i : snk_i]) 
				bubbles_from_src[src_seq].add(SrcSnk_Seq)

				try:
					SRCSNK_pairs[Read_Index][(src_seq, snk_seq)].add(SrcSnk_Seq)
				except:
					SRCSNK_pairs[Read_Index][(src_seq, snk_seq)] = set([SrcSnk_Seq])

	paths_per_src = [len(v) for v in bubbles_from_src.values()]

	complexity = np.sum(paths_per_src)


	if complexity >= 5e3:	
		SRCSNK_pairs = {}
	
	return SRCSNK_pairs



class seq_interval_obj():
	def __init__(self, src, start, end, dummy_var = None):
		self.src = src
		if int(end) < int(start):
			raise ValueError() 
		self.start = int(start)
		self.end   = int(end)
		self.repeat = 0

	def __len__(self):
		return self.end - self.start 

	def get_seq(self, reference_sequence):
		return reference_sequence[self.start : self.end].upper()

	def add_repeat(self):
		self.repeat += 1

	def __str__(self):
		return '{0.src}|{0.start}|{0.end}|{0.repeat}'.format(self)





def make_read_dbg(Read_Index, Ref_Index, SetOfReadIndexes):

	EVENTS = SetOfReadIndexes[Read_Index].Events
	Rid_Long_Seq = SetOfReadIndexes[Read_Index].LongSeq
	Ref_Long_Seq = SetOfReadIndexes[Ref_Index].LongSeq
	Rid_Long_Seq_ext = "YY" + Rid_Long_Seq + "YY" 

	Graph = nx.DiGraph()

	rid_nodes = set([0 , 2, len(Rid_Long_Seq_ext) - 2, len(Rid_Long_Seq_ext)])
	events_n  = 0
	events_skipped = 0

	for Rid_iv in EVENTS:
		rid_nodes.add(Rid_iv.start + 2)
		rid_nodes.add(Rid_iv.end + 2)
		events_n += len(EVENTS[Rid_iv])

	rid_nodes = list(rid_nodes)
	rid_nodes.sort()

	for i in range(len(rid_nodes) - 2):
		i1, i2, i3 = rid_nodes[i : i + 3]
		node_src = seq_interval_obj("RID", i1, i2)
		node_snk = seq_interval_obj("RID", i2, i3) 
		Graph.add_edge(str(node_src), str(node_snk), label = "READSEQ")
		
	for Rid_iv in EVENTS:
		r_i1 = rid_nodes.index(Rid_iv.start + 2)
		r_i2 = rid_nodes.index(Rid_iv.end + 2, r_i1)

		node_src = seq_interval_obj("RID", rid_nodes[r_i1 - 1], rid_nodes[r_i1])
		node_snk = seq_interval_obj("RID", rid_nodes[r_i2], rid_nodes[r_i2 + 1])
		
		for Ref_iv in EVENTS[Rid_iv]:

			s1, s2 = Rid_iv.get_seq(Rid_Long_Seq), Ref_iv.get_seq(Ref_Long_Seq)

			if s1 == s2:
				continue
			
			while str(Ref_iv) in Graph.nodes:
				Ref_iv.add_repeat() 

			if events_n >= 500 and ((s1 in s2) or (s2 in s1)):
				events_skipped += 1
				continue 

			if events_n >= 500 and abs(len(Ref_iv) - (node_snk.start - node_src.end)) > 5:
				events_skipped += 1
				continue 
			
			Graph.add_edge(str(node_src), str(Ref_iv), label = "REFSEQ")
			Graph.add_edge(str(Ref_iv), str(node_snk), label = "REFSEQ")

	source = seq_interval_obj("RID", rid_nodes[0], rid_nodes[1])
	target = seq_interval_obj("RID", rid_nodes[-2], rid_nodes[-1])

	return Graph, str(source), str(target)



def Dijkstra_find_end_to_end_paths(Read_Index, Ref_Index, SetOfReadIndexes, Graph, 
		source, target, edges, REFPOS):

	Ref_Long_Seq = SetOfReadIndexes[Ref_Index].LongSeq
	Rid_Long_Seq = SetOfReadIndexes[Read_Index].LongSeq

	path = dict()
	Rid_Long_Seq_ext = "YY" + Rid_Long_Seq + "YY"

	for node in Graph.nodes:
		node_obj = seq_interval_obj(*node.split('|'))
		if node_obj.src == "RID":
			node_obj_seq = node_obj.get_seq(Rid_Long_Seq_ext)
		else:
			node_obj_seq = node_obj.get_seq(Ref_Long_Seq)

		path[node] = {'SEQ' : node_obj_seq.strip('Y'), 'PATH' : {}}

	if not path:
		return {}

	path[source]['PATH'][source] = {'ref_pos' : None, 'path_seq' : path[source]['SEQ']}

	queue = set([(source, 0)])
	
	stack_counter = 0

	while queue:

		(u, up) = min(queue, key = lambda x: x[1])
		queue.remove((u, up))

		u_obj = seq_interval_obj(*u.split("|"))

		ref_pos_set = set([vv['ref_pos'] for p, vv in path[u]['PATH'].items() ])
		
		ref_pos_set_counter = defaultdict(int)
		if len(path[u]['PATH']) * Graph.out_degree(u) > 2e5:
			for u_path, u_attr in list(path[u]['PATH'].items()):
				ref_pos = u_attr['ref_pos']
				if u_path.count("REF") == 0:
					continue
				elif ref_pos_set_counter[ref_pos] >= 5:
					path[u]['PATH'].pop(u_path)
				else:
					ref_pos_set_counter[ref_pos] += 1 


		for v in Graph.successors(u):

			v_obj = seq_interval_obj(*v.split("|"))
			v_obj_Seq = path[v]['SEQ']

			for u_path, u_attr in path[u]['PATH'].items():
				stack_counter += 1

				path_coord    = u_path + ';' + v 
				path_sequence = u_attr['path_seq'] + v_obj_Seq
				ref_pos       = u_attr['ref_pos']

				if v_obj.src == "REF" and Ref_Long_Seq[:v_obj.end].endswith(path_sequence):

					tmp_ref_pos = v_obj.end - len(path_sequence)

					if ref_pos is not None and tmp_ref_pos != ref_pos:
						continue
			
					w = Graph.successors(v).__next__()
					
					w_obj = seq_interval_obj(*w.split("|"))
					w_obj_Seq = path[w]['SEQ']
					
					assert Graph.out_degree(v) == 1 and Graph.in_degree(v) == 1
					assert w_obj.src == "RID" and u_obj.src == "RID"

					path_coord_ext    = u_path + ';' + v + ';' + w
					path_sequence_ext = u_attr['path_seq'] + v_obj_Seq + w_obj_Seq

					if Ref_Long_Seq[tmp_ref_pos:].startswith(path_sequence_ext):
						if path_coord_ext.count('REF') > 5:
							continue
						path[w]['PATH'][path_coord_ext] = {'ref_pos'  : tmp_ref_pos, 
														   'path_seq' : path_sequence_ext}
						
						edges[(v, w)].append('REF')
						queue.add((w, w_obj.start))
						
				elif v_obj.src == "RID":
					if ref_pos is None or Ref_Long_Seq[ref_pos:].startswith(path_sequence):
						path[v]['PATH'][path_coord] = {'ref_pos'  : ref_pos, 
													   'path_seq' : path_sequence}
						queue.add((v, v_obj.start))

		if u != target:
			path.pop(u)

		if stack_counter > int(2e5):
			return None
	 
	for _path, _attr in list(path[target]['PATH'].items()):
		pos, seq = _attr['ref_pos'], _attr['path_seq']
		if pos is None:
			if "REF" not in _path and Rid_Long_Seq in Ref_Long_Seq:
				pos = Ref_Long_Seq.index(Rid_Long_Seq)
				path[target]['PATH'][_path]['ref_pos'] = pos
			else:
				path[target]['PATH'].pop(_path)
		else:
			assert Ref_Long_Seq[pos : pos + len(seq)] == seq

	start_to_end_paths = dict((p, path[target]['PATH'][p]['ref_pos']) for p in path[target]['PATH'])
	
	SOFTMAN.rm_nested_dict(path) 

	return start_to_end_paths



def _dbg_compress_x(nd, k):
	if len(nd) > 2*k:
		return nd[:k] + '_' + nd[-k:]
	else:
		return nd



def draw_rid_graph(xdb, graph_file_name):
	out = open("{}.txt".format(graph_file_name), "w")
	out.write('digraph G {\n')
	out.write('size="7,7"\nrankdir=TB;\n')

	for nd in xdb.nodes:
		nd = nd.replace("|","_")
		if nd.startswith("RID"):
			color = "red"
		else:
			color = "blue"
		string1 = '\t{0} [color={1},label={0}];\n'.format(nd, color)
		out.write(string1)

	for nd_or, nd_dt in xdb.edges:
		nd_or = nd_or.replace("|","_")
		nd_dt = nd_dt.replace("|","_")
		label = ''
		if nd_or.startswith("RID") and nd_dt.startswith("RID"):
			color = "red"
		else:
			color = "black"

		string = '\t{0} -> {1} [color={3},label="{2}"];\n'.format(nd_or, nd_dt, label, color)
		out.write(string)

	out.write('}')
	out.close()


def draw_de_Bruijn_graph(xdb, k, graph_file_name, Ref_Index = 0):
	out = open("{}.txt".format(graph_file_name), "w")
	out.write('digraph G {\n')
	out.write('size="15,15"\nrankdir=TB;\n')

	for nd in xdb.nodes:
		nd_str  = _dbg_compress_x(nd, k)
		string1 = '\t{0} [label={0}_{1}];\n'.format(nd_str, len(nd) - k + 2)
		out.write(string1 )

	for nd_or, nd_dt in xdb.edges:
		N = xdb[nd_or][nd_dt]['n']

		label = "{}".format(N)
		color = "blue"

		for x in xdb[nd_or][nd_dt]['reads']:
			if type(x) is tuple and x[0] == Ref_Index :
				label = "{}_REF".format(N)
				color = "red"
				break
			elif type(x) is int and x == Ref_Index:
				label = "{}_REF".format(N)
				color = "red"
				break
		# if len(xdb[nd_or][nd_dt]['reads']) == 0:
		# 	continue

		nd_or = _dbg_compress_x(nd_or, k)
		nd_dt = _dbg_compress_x(nd_dt, k)

		string = '\t{0} -> {1} [color={3},label="{2}"];\n'.format(nd_or, nd_dt, label, color)
		out.write(string)

	out.write('}')
	out.close()
